# Generated by Django 3.0.3 on 2020-08-25 11:35

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0005_auto_20200825_1128'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='labels',
            name='complete',
        ),
    ]



# The MIT License (MIT)
#
# Copyright (c) 2015 University of East Anglia, Norwich, UK
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Developed by Geoffrey French in collaboration with Dr. M. Fisher and
# Dr. M. Mackiewicz.
from typing import Any
from abc import abstractmethod
import uuid
import copy
import numpy as np
from PyQt5 import QtCore, QtWebChannel
from image_labelling_tool import labelling_tool, labelling_schema, schema_editor_messages
from image_labelling_tool.flask_labeller import image_url_salt
from image_labelling_tool_qt import web_server


def _json_to_qt(js):
    """
    Internal helper function

    Qt is supposed to coerce between QJsonValue types and python JSON types transparently.
    This mostly works, except if your JSON data contains None/null. In that case Qt fails
    to convert Python types to Qt JSON types, so we define this function to do it by hand.

    :param js: JSON data (int/float/bool/list/dict/None)
    :return: QtCore.QJsonValue instance
    """
    if js is None:
        return QtCore.QJsonValue()
    elif isinstance(js, (bool, int, float, str)):
        return QtCore.QJsonValue(js)
    elif isinstance(js, list):
        return QtCore.QJsonValue([_json_to_qt(v) for v in js])
    elif isinstance(js, dict):
        return QtCore.QJsonValue({k: _json_to_qt(v) for k, v in js.items()})
    else:
        raise TypeError('Cant handle type {}'.format(type(js)))


def _qt_to_json(qtjs):
    """
    Internal helper function

    We handle QJsonValue to Python JSON types manually to paper over other issues...
    :param qtjs: a `QJsonValue` instance
    :return: JSON data (int/float/bool/list/dict/None)
    """
    if qtjs.isBool():
        return qtjs.toBool()
    elif qtjs.isDouble():
        return qtjs.toDouble()
    elif qtjs.isNull() or qtjs.isUndefined():
        return None
    elif qtjs.isString():
        return str(qtjs.toString())
    elif qtjs.isArray():
        arr = qtjs.toArray()
        return [_qt_to_json(x) for x in arr]
    elif qtjs.isObject():
        obj = qtjs.toObject()
        return {str(k): _qt_to_json(obj.get(k)) for k in obj.keys()}
    else:
        raise RuntimeError


class QAbstractDLTool (QtCore.QObject):
    """Qt side tool that is to be attached to a `QtWebEngineWidgets.QWebEngineView` widget.

    WARNING: You must ensure that a reference to this object (`self`) is kept around/alive as long
    as it is needed for display in a web engine view. As of Qt 5.9.7 and PyQt 5.9.2 (the versions
    available on Anaconda that this was tested on in Mar/2021), the call to `QWebChannel.registerObject`
    that is made in the `attach_to_web_engine_view` method does *not* seem to prevent Python's
    garbage collector from deleting `self`. As a consequence, it is important to ensure that
    a reference to `self` is kept as long as it is needed for display.
    Currently the `attach_to_web_engine_view` assist this by assigning a reference to `self`
    to the `_django_labeller` attribute of the `QtWebEngineWidgets.QWebEngineView` object
    passed as a parameter to `attach_to_web_engine_view`. If you need `self` to last longer,
    please keep a reference to it.
    """
    _TOOL_URL_NAME = None

    def __init__(self, server):
        """
        :param server: the `web_server.LabellerServer` instance that manages the Flask server.
        """
        super(QAbstractDLTool, self).__init__()

        self._server = server
        self._server_pipe = self._server.image_registry()

        self._tool_id = str(uuid.uuid4())

    def attach_to_web_engine_view(self, web_engine_view):
        """
        Attach the labeller to a `QtWebEngineWidgets.QWebEngineView` widget

        WARNING: You must ensure that a reference to this object (`self`) is kept around/alive as long
        as it is needed for display in a web engine view. As of Qt 5.9.7 and PyQt 5.9.2 (the versions
        available on Anaconda that this was tested on in Mar/2021), the call to `QWebChannel.registerObject`
        that is made in the `attach_to_web_engine_view` method does *not* seem to prevent Python's
        garbage collector from deleting `self`. As a consequence, it is important to ensure that
        a reference to `self` is kept as long as it is needed for display.
        Currently the `attach_to_web_engine_view` assist this by assigning a reference to `self`
        to the `_django_labeller` attribute of the `QtWebEngineWidgets.QWebEngineView` object
        passed as a parameter. If you need `self` to last longer, please keep a reference to it.

        :param web_engine_view: A `QtWebEngineWidgets.QWebEngineView` instance that is the widget in which the
            labeller is to be rendered
        """
        if self._TOOL_URL_NAME is None:
            raise NotImplementedError('Abstract: the _TOOL_URL_NAME class attribute is not defined for {}'.format(
                type(self)
            ))
        # Start the Flask server if its not already running
        self._server.start_flask_server()
        # Create a QWebChannel to communicate with the client-side Javascript code
        channel = QtWebChannel.QWebChannel(web_engine_view.page())
        web_engine_view.page().setWebChannel(channel)
        # Register self as `qt_tool` as the `labeller_control_qt.jinja2` template will attempte to find it here
        channel.registerObject("qt_tool", self)
        # Get the server url
        tool_url = self._server.server_url(self._TOOL_URL_NAME, tool_id=self._tool_id,
                                           query_params=self.get_server_query_params())
        # Have the web view widget navigate there
        web_engine_view.setUrl(QtCore.QUrl(tool_url))
        # Set the '_django_labeller' attribute so that self is not garbage collected while
        # `web_engine_view` still exists
        web_engine_view._django_labeller = self

    def get_server_query_params(self):
        """Return any additional GET parameters that should be added to the flask server URL"""
        return {}


class QAbstractLabeller (QAbstractDLTool):
    """Qt side labeller that is to be attached to a `QtWebEngineWidgets.QWebEngineView` widget.

    Concrete implementations should implement the following methods and properties:

    image_descriptors_json: get the list of images that the tool should present to the user
    get_image_labels_for_tool(): retrieve the image labels for a given image to be displayed by the tool
    on_update_image_labels_from_tool(): update the labels for a given image
    dextr_predict_mask(): predict a mask using DEXTR for automatically assisted labelling
    dextr_available: determine if DEXTR is available/supported

    WARNING: You must ensure that a reference to this object (`self`) is kept around/alive as long
    as it is needed for display in a web engine view. As of Qt 5.9.7 and PyQt 5.9.2 (the versions
    available on Anaconda that this was tested on in Mar/2021), the call to `QWebChannel.registerObject`
    that is made in the `attach_to_web_engine_view` method does *not* seem to prevent Python's
    garbage collector from deleting `self`. As a consequence, it is important to ensure that
    a reference to `self` is kept as long as it is needed for display.
    Currently the `attach_to_web_engine_view` assist this by assigning a reference to `self`
    to the `_django_labeller` attribute of the `QtWebEngineWidgets.QWebEngineView` object
    passed as a parameter to `attach_to_web_engine_view`. If you need `self` to last longer,
    please keep a reference to it.
    """
    _TOOL_URL_NAME = 'labeller'

    def __init__(self, server, schema, tasks=None,
                 anno_controls=None, config=None, enable_firebug=False):
        """
        :param server: the `web_server.LabellerServer` instance that manages the Flask server.
        :param schema: labelling schema that will be passed to the tool
        :param tasks: [optional] a list of tasks for the user to check when they are done
        :param anno_controls: [optional] additional annotation controls for metadata
        :param config: [optional] labelling tool configuration
        :param enable_firebug: [default=False] if True, load Firebug-lite development tools
        """
        super(QAbstractLabeller, self).__init__(server)

        self._dextr_fn = None

        if tasks is None:
            tasks = []

        # Convert the schema to JSON
        if isinstance(schema, labelling_schema.LabellingSchema):
            self._schema_json = schema.to_json()
        elif isinstance(schema, labelling_schema.SchemaStore):
            self._schema_json = schema.get_schema_json()
        else:
            self._schema_json = schema

        # Create a default 'finished' task if none provided
        self._tasks = tasks

        # Annotation metadata controls
        if anno_controls is not None:
            self._anno_controls_json = [c.to_json() for c in anno_controls]
        else:
            self._anno_controls_json = []

        # Configuration
        if config is None:
            # None provided, use default
            config = labelling_tool.DEFAULT_CONFIG
        # Copy it
        config = copy.deepcopy(config)
        # Set settings.fullscreenButton to False if not provided
        config.setdefault('settings', dict()).setdefault('fullscreenButton', False)
        self._config = config

        settings = dict(
            config=config,
            tasks=tasks,
            schema=self._schema_json,
            anno_controls=self._anno_controls_json,
            enable_firebug=enable_firebug,
        )
        self._server_pipe.add_settings(self._tool_id, settings)

    def get_server_query_params(self):
        if self._dextr_fn is not None:
            return {'dextr': ''}
        else:
            return {}

    @property
    @abstractmethod
    def image_descriptors_json(self):
        """
        A list of JSON image descriptors that will be passed to the client side tool that describe the images that
        are available for labelling.

        Create a descriptor for each image by invoking:
        `descr = labelling_tool.image_descriptor(image_id=image_id, url=url, width=width, height=height)`
        Note that `url` can be None, in which case it will be filled in for you,
        otherwise it should be the URL at which the image can be found on the Flask server.
        The URL would be '/image/<image_id>'

        :return: a list of JSON image descriptors
        """
        pass

    @abstractmethod
    def get_image_labels_for_tool(self, image_id):
        """
        Retrieve the image labels and metadata for use by the tool.

        :param image_id: the Image ID identifying the image
        :return: a dictionary with the following layout:
            labels_json: labels in JSON format
            completed_tasks: the list of names of tasks that have been completed by the user as a list of strings
        """
        pass

    @abstractmethod
    def on_update_image_labels_from_tool(self, image_id, labels_and_metadata):
        """
        Invoked when the tool supplies labels modified by the user

        :param image_id: the Image ID that identifies the image whose labels we are updating
        :param labels_and_metadata: a dictionary with the following layout:
            labels_json: updated labels in JSON format
            completed_tasks: the list of names of tasks that have been completed by the user as a list of strings
        """
        pass

    @abstractmethod
    def dextr_predict_mask(self, image_id, dextr_points):
        """
        Use DEXTR to predict a mask for an object that is identified by the points in `dextr_points`
        :param image_id: the image ID identifying the image that should be used for inference
        :param dextr_points: points to identify the image as a `(N, [y, x])` NumPy array
        :return: a mask as a `(H, W)` NumPy array whose size matches that of the image
        """
        pass

    @property
    @abstractmethod
    def dextr_available(self):
        """Return True if DEXTR automatically assisted labelling is available"""
        pass


    _tool_load_labels = QtCore.pyqtSignal("QJsonObject")
    _tool_set_labels_reply = QtCore.pyqtSignal("QJsonObject")
    _tool_dextr_reply = QtCore.pyqtSignal("QJsonObject")

    @QtCore.pyqtSlot(str)
    def _tool_request_labels(self, image_id_str):
        image_id_str = str(image_id_str)
        labels_and_metadata = self.get_image_labels_for_tool(image_id_str)
        label_header = dict(
            image_id=image_id_str,
            labels=_json_to_qt(labels_and_metadata['labels_json']),
            completed_tasks=labels_and_metadata['completed_tasks'],
            timeElapsed=0.0,
            state='editable',
            session_id=str(uuid.uuid4()),
        )
        self._tool_load_labels.emit(label_header)

    @QtCore.pyqtSlot("QJsonValue")
    def _tool_set_labels(self, label_header):
        label_header = _qt_to_json(label_header)
        image_id_str = label_header['image_id']
        labels_and_metadata = dict(labels_json=label_header['labels'],
                                   completed_tasks=label_header['completed_tasks'])
        self.on_update_image_labels_from_tool(image_id_str, labels_and_metadata)
        self._tool_set_labels_reply.emit({})

    @QtCore.pyqtSlot("QJsonValue")
    def _tool_dextr(self, dextr_js):
        dextr_js = _qt_to_json(dextr_js)
        if 'request' in dextr_js:
            dextr_request_js = dextr_js['request']
            image_id = dextr_request_js['image_id']
            dextr_id = dextr_request_js['dextr_id']
            dextr_points = dextr_request_js['dextr_points']

            if self.dextr_available:
                # Convert to `[N, [y, x]]` NumPy array
                dextr_points_np = np.array([[p['y'], p['x']] for p in dextr_points])
                # Predict mask
                mask = self.dextr_predict_mask(image_id, dextr_points_np)
                # Convert to vector label
                regions = labelling_tool.PolygonLabel.mask_image_to_regions_cv(mask, sort_decreasing_area=True)
                # To JSON
                regions_js = labelling_tool.PolygonLabel.regions_to_json(regions)
            else:
                regions_js = []

            dextr_labels = dict(image_id=image_id, dextr_id=dextr_id, regions=regions_js)
            dextr_reply = dict(labels=[dextr_labels])
            self._tool_dextr_reply.emit(dextr_reply)
        elif 'poll' in dextr_js:
            dextr_reply = dict(labels=[])
            self._tool_dextr_reply.emit(dextr_reply)
        else:
            raise RuntimeError

    @QtCore.pyqtProperty("QJsonObject")
    def _tool_schema(self):
        return dict(value=self._schema_json)

    @QtCore.pyqtProperty("QJsonObject")
    def _tool_anno_tasks(self):
        return dict(value=self._tasks)

    @QtCore.pyqtProperty("QJsonObject")
    def _tool_anno_controls(self):
        return dict(value=self._anno_controls_json)

    @QtCore.pyqtProperty("QJsonObject")
    def _tool_image_descriptors(self):
        descrs = self.image_descriptors_json.copy()
        # Fixup the URLs
        for d in descrs:
            if d['img_url'] is None:
                d['img_url'] = '/image/{}'.format(d['image_id'])
        return dict(value=descrs)

    @QtCore.pyqtProperty("QJsonObject")
    def _tool_anno_config(self):
        return dict(value=self._config)

    @QtCore.pyqtProperty(bool)
    def _tool_dextr_available(self):
        return self.dextr_available


class QLabellerForLabelledImages (QAbstractLabeller):
    """Qt side labeller that is to be attached to a `QtWebEngineWidgets.QWebEngineView` widget.

    Operates on a provided list of labelled images (see `labelled_image.LabelledImage`)

    WARNING: You must ensure that a reference to this object (`self`) is kept around/alive as long
    as it is needed for display in a web engine view. As of Qt 5.9.7 and PyQt 5.9.2 (the versions
    available on Anaconda that this was tested on in Mar/2021), the call to `QWebChannel.registerObject`
    that is made in the `attach_to_web_engine_view` method does *not* seem to prevent Python's
    garbage collector from deleting `self`. As a consequence, it is important to ensure that
    a reference to `self` is kept as long as it is needed for display.
    Currently the `attach_to_web_engine_view` assist this by assigning a reference to `self`
    to the `_django_labeller` attribute of the `QtWebEngineWidgets.QWebEngineView` object
    passed as a parameter to `attach_to_web_engine_view`. If you need `self` to last longer,
    please keep a reference to it.
    """
    def __init__(self, server, labelled_images, schema, tasks=None, colour_schemes=None,
                 anno_controls=None, config=None, dextr_fn=None, enable_firebug=False):
        """
        :param server: the `web_server.LabellerServer` instance that manages the Flask server.
        :param label_classes: grouped label classes that will be passed to the tool
        :param labelled_images: a list of labelled images to edit (see `labelling_tool.AbstractLabelledImage`)
        :param tasks: [optional] a list of tasks for the user to check when they are done
        :param colour_schemes: [optional] the list of colour schemes for display
        :param anno_controls: [optional] additional annotation controls for metadata
        :param config: [optional] labelling tool configuration
        :param dextr_fn: [optional] DEXTR prediction function
        :param enable_firebug: [default=False] if True, load Firebug-lite development tools
        """
        super(QLabellerForLabelledImages, self).__init__(
            server=server, schema=schema, tasks=tasks, anno_controls=anno_controls,
            config=config, enable_firebug=enable_firebug)

        self._dextr_fn = dextr_fn

        # Generate image IDs list
        image_ids = ['{}__{}'.format(self._tool_id, i) for i in range(len(labelled_images))]
        # Generate images table mapping image ID to image so we can get an image by ID
        images_table = {image_id: img for image_id, img in zip(image_ids, labelled_images)}
        # Generate image descriptors list to hand over to the labelling tool
        # Each descriptor provides the image ID, the URL and the size
        self.__image_descriptors = []
        for image_id, img in zip(image_ids, labelled_images):
            height, width = img.image_source.image_size
            local_path = img.image_source.local_path
            if local_path is not None:
                self._server_pipe.add_image(image_id, web_server._ImagePath(
                    path=str(local_path.absolute()), width=width, height=height))
            else:
                data, mime_type = img.image_source.image_binary_and_mime_type()
                self._server_pipe.add_image(image_id, web_server._ImageBinary(
                    data=data, mime_type=mime_type, width=width, height=height))
            # The 'salt' added to the URL prevents the browser cache from displaying images cached from
            # a previous session
            self.__image_descriptors.append(labelling_tool.image_descriptor(
                image_id=image_id, url='/image/{}?salt={}'.format(image_id, image_url_salt()),
                width=width, height=height
            ))
        self.__images_table = images_table

    @property
    def image_descriptors_json(self):
        return self.__image_descriptors

    def get_image_labels_for_tool(self, image_id):
        image = self.__images_table[image_id]
        wrapped_labels = image.labels_store.get_wrapped_labels()
        return dict(labels_json=wrapped_labels.labels_json, completed_tasks=wrapped_labels.completed_tasks)

    def on_update_image_labels_from_tool(self, image_id, labels_and_metadata):
        image = self.__images_table[image_id]

        wrapped_labels = image.labels_store.get_wrapped_labels()
        wrapped_labels.labels_json = labels_and_metadata['labels_json']
        wrapped_labels.completed_tasks = labels_and_metadata['completed_tasks']
        image.labels_store.update_wrapped_labels(wrapped_labels)

    def dextr_predict_mask(self, image_id, dextr_points):
        image = self.__images_table[image_id]
        return self._dextr_fn(image.image_source.image_as_array_or_pil(), dextr_points)

    @property
    def dextr_available(self):
        return self._dextr_fn is not None


class QAbstractSchemaEditor (QAbstractDLTool, schema_editor_messages.SchemaEditorMessageHandler):
    """Qt side schema editor that is to be attached to a `QtWebEngineWidgets.QWebEngineView` widget.

    Concrete implementations should implement the following methods:

    get_schema_json(): retrieve the schema in JSON form in its current state
    update_schema_json(schema_js): update schema in JSON form

    WARNING: You must ensure that a reference to this object (`self`) is kept around/alive as long
    as it is needed for display in a web engine view. As of Qt 5.9.7 and PyQt 5.9.2 (the versions
    available on Anaconda that this was tested on in Mar/2021), the call to `QWebChannel.registerObject`
    that is made in the `attach_to_web_engine_view` method does *not* seem to prevent Python's
    garbage collector from deleting `self`. As a consequence, it is important to ensure that
    a reference to `self` is kept as long as it is needed for display.
    Currently the `attach_to_web_engine_view` assist this by assigning a reference to `self`
    to the `_django_labeller` attribute of the `QtWebEngineWidgets.QWebEngineView` object
    passed as a parameter to `attach_to_web_engine_view`. If you need `self` to last longer,
    please keep a reference to it.
    """
    _TOOL_URL_NAME = 'schema_editor'

    def __init__(self, server, enable_firebug=False):
        """
        :param server: the `web_server.LabellerServer` instance that manages the Flask server.
        :param schema_json: schema in JSON form
        :param enable_firebug: [default=False] if True, load Firebug-lite development tools
        """
        super(QAbstractSchemaEditor, self).__init__(server)
        
        settings = dict(
            schema=self.get_schema_json(),
            enable_firebug=enable_firebug,
        )
        self._server_pipe.add_settings(self._tool_id, settings)

    @abstractmethod
    def get_schema_json(self):
        pass

    @abstractmethod
    def update_schema_json(self, schema_js):
        pass

    # Schema editor message handler methods
    def update_schema(self, request, schema, schema_js: Any):
        self.update_schema_json(schema_js)

    def create_colour_scheme(self, request, schema, colour_scheme_js: Any):
        return None

    def delete_colour_scheme(self, request, schema, colour_scheme_js: Any):
        pass

    def create_group(self, request, schema, group_js: Any):
        return None

    def delete_group(self, request, schema, group_js: Any):
        pass

    def create_label_class(self, request, schema, containing_group_js: Any, label_class_js: Any):
        return None

    def delete_label_class(self, request, schema, containing_group_js: Any, label_class_js: Any):
        pass

    _notify_response = QtCore.pyqtSignal("QJsonObject")

    @QtCore.pyqtSlot("QJsonValue")
    def _update_callback(self, message_block):
        message_block = _qt_to_json(message_block)
        responses = self.handle_messages(None, None, message_block['messages'])
        # Copy the message block ID field over so that the client knows which message block
        # we are responding to
        responses['id'] = message_block['id']
        self._notify_response.emit(responses)

    @QtCore.pyqtProperty("QJsonObject")
    def _schema(self):
        return dict(value=self.get_schema_json())


class QSchemaEditorForSchemaStore (QAbstractSchemaEditor):
    """Qt side labeller that is to be attached to a `QtWebEngineWidgets.QWebEngineView` widget.

    Operates on a provided `labelling_schema.SchemaStore` instance

    WARNING: You must ensure that a reference to this object (`self`) is kept around/alive as long
    as it is needed for display in a web engine view. As of Qt 5.9.7 and PyQt 5.9.2 (the versions
    available on Anaconda that this was tested on in Mar/2021), the call to `QWebChannel.registerObject`
    that is made in the `attach_to_web_engine_view` method does *not* seem to prevent Python's
    garbage collector from deleting `self`. As a consequence, it is important to ensure that
    a reference to `self` is kept as long as it is needed for display.
    Currently the `attach_to_web_engine_view` assist this by assigning a reference to `self`
    to the `_django_labeller` attribute of the `QtWebEngineWidgets.QWebEngineView` object
    passed as a parameter to `attach_to_web_engine_view`. If you need `self` to last longer,
    please keep a reference to it.
    """
    def __init__(self, server, schema_store: labelling_schema.SchemaStore, enable_firebug=False):
        """
        :param server: the `web_server.LabellerServer` instance that manages the Flask server.
        :param schema_store: a `labelling_schema.SchemaStore` instance
        :param enable_firebug: [default=False] if True, load Firebug-lite development tools
        """
        self._schema_store = schema_store

        super(QSchemaEditorForSchemaStore, self).__init__(
            server=server, enable_firebug=enable_firebug)

    def get_schema_json(self):
        return self._schema_store.get_schema_json()

    def update_schema_json(self, schema_js):
        self._schema_store.update_schema_json(schema_js)




from django.db import models

from image_labelling_tool import models as lt_models
import django.utils.timezone

# Create your models here.
class ImageWithLabels (models.Model):
    # image
    image = models.ImageField(blank=True)

    # labels
    labels = models.ForeignKey(lt_models.Labels, models.CASCADE, related_name='image')


class DextrTask (models.Model):
    creation_timestamp = models.DateTimeField(default=django.utils.timezone.now)
    image = models.ForeignKey(ImageWithLabels, models.CASCADE)
    image_id_str = models.CharField(max_length=128)
    dextr_id = models.IntegerField()
    celery_task_id = models.CharField(max_length=128)



from django.contrib import admin
from . import models

# Register your models here.
admin.site.register(models.LabellingTask)
admin.site.register(models.LabellingSchema)
admin.site.register(models.LabellingColourScheme)
admin.site.register(models.LabelClassGroup)
admin.site.register(models.LabelClass)
admin.site.register(models.LabelClassColour)
admin.site.register(models.Labels)



# The MIT License (MIT)
#
# Copyright (c) 2015 University of East Anglia, Norwich, UK
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Developed by Geoffrey French in collaboration with Dr. M. Fisher and
# Dr. M. Mackiewicz.
import json
from abc import abstractmethod
import io
import math
import itertools
import pathlib
import re
from typing import Any, Optional, Union, Container, Sequence, Tuple, List, Generator
from typing import Mapping, MutableMapping, Dict, Callable, IO
import copy
from deprecated import deprecated

import numpy as np

import uuid

from PIL import Image, ImageDraw

from skimage.color import gray2rgb
from skimage.measure import find_contours

from image_labelling_tool.labelling_schema import LabelClass, LabelClassGroup, ColourTriple

# Try to import cv2
try:
    import cv2
except:
    cv2 = None


# Configuration is composed of nested dictionaries. It's structure is given below.
# If values or sections are not present, then the defaults specified below will be used.
#
# config: {
#     tools: {
#         imageSelector: bool [default=True] if True, show the UI that allows the user to switch between images
#         labelClassSelector: bool [default=True] if True, show the label class selection UI and allow the user to
#                 assign classes to labels
#         labelClassFilter: <bool> [default=False] if True, allow the user to choose to view only labels belonging to
#                 a specific class
#         labelClassFilterInitial: <string|False|None> [default=None] the initial label class filter; at startup
#                 only labels of this class will be visible. If given a value of None, only labels that
#                 have no assigned class will be shown. If given a value of False, all labels will be shown.
#         brushSelect: <bool> [default=True] if True, enable the brush select tool
#         drawPointLabel: <bool> [default=True] if True, enable the create/draw point label tool
#         drawBoxLabel: <bool> [default=True] if True, enable the create/draw box label tool
#         drawOrientedEllipseLabel: <bool> [default=True] if True, enable the create/drawn oriented ellipse label tool
#         drawPolyLabel: <bool> [default=True] if True, enable the create/drawn polygonal label tool
#         groupLabel: <bool> [default=True] if True, enable the create group label tool
#         deleteLabel: <bool> [default=True] if True, enable the delete label tool
#         legacy: {
#             compositeLabel: <bool> [default=True] if True, enable the LEGACY create composite label tool
#         }
#         deleteConfig: {
#             typePermissions: {
#                 point: <bool> [default=True] if True, allow the user to delete point labels
#                 box: <bool> [default=True] if True, allow the user to delete box labels
#                 oriented_ellipse: <bool> [default=True] if True, allow the user to delete oriented ellipse labels
#                 polygon: <bool> [default=True] if True, allow the user to delete polygonal labels
#                 composite: <bool> [default=True] if True, allow the user to delete LEGACY composite labels
#                 group: <bool> [default=True] if True, allow the user to delete group labels (deleting a group
#                         label will ungroup the child labels, leaving them in place
#             }
#         }
#         nextUnlockedConfig: {
#             numImagesLimit: <int> [default=100] when using locking in a multi-user app (e.g. in a Django app),
#                     this is the maximum number of images that will be searched for the next available unlocked image
#         }
#     }
#     settings: {
#         inactivityTimeoutMS: <int> [default=10000] the period of inactivity after which the tool will stop
#                 accumulating time in the images' timeElapsed metadata that measures the amount of time users have
#                 spend annotating an current image
#         brushWheelRate: <float> [default=0.025] the scale factor that affects the rate at which the mouse
#                 wheel changes the brush size (measured in image pixels) when using the brush select tool
#         brushKeyRate: <float> [default=2.0] the amount by which keyboard shortcuts modify (up or down) the
#                 brush size (measured in image pixels) when using the brush select tool
#         fullscreenButton: <bool> [default=True] if True, enable the full screen button
#     }
# }
DEFAULT_CONFIG = {
    'tools': {
        'imageSelector': True,
        'labelClassSelector': True,
        'drawPointLabel': False,
        'drawBoxLabel': True,
        'drawOrientedEllipseLabel': True,
        'drawPolyLabel': True,
        'deleteLabel': True,
        'deleteConfig': {
            'typePermissions': {
                'point': True,
                'box': True,
                'polygon': True,
                'composite': True,
                'group': True,
            }
        },
        'legacy': {
            'compositeLabel': False
        }
    },
    'settings': {
        'brushWheelRate': 0.025,  # Change rate for brush radius (mouse wheel)
        'brushKeyRate': 2.0,  # Change rate for brush radius (keyboard)
        'fullscreenButton': False,
    }
}


@deprecated(reason='Please use LabelClass in the labelling_schema module')
def label_class(name: str, human_name: str, rgb: Tuple[int, int, int]) -> Any:
    return {'name': name,
            'human_name': human_name,
            'colours': {'default': rgb}}

@deprecated(reason='Please use LabelClassGroup in the labelling_schema module')
def label_class_group(human_name: str, classes_json: List[Any]) -> Any:
    return {'group_name': human_name,
            'group_classes': classes_json}

def image_descriptor(image_id: Any, url: Optional[Any] = None,
                     width: Optional[int] = None, height: Optional[int] = None):
    return {'image_id': str(image_id),
            'img_url': str(url) if url is not None else None,
            'width': width,
            'height': height,}


class _AnnoControl (object):
    __control_type__ = None

    def __init__(self, identifier: str):
        self.identifier = identifier

    def to_json(self) -> Any:
        return dict(control=self.__control_type__, identifier=self.identifier)


class _AnnoControlVis (_AnnoControl):
    __control_type__ = None

    def __init__(self, identifier: str, visibility_label_text: Optional[str] = None):
        super(_AnnoControlVis, self).__init__(identifier)
        self.visibility_label_text = visibility_label_text

    def to_json(self) -> Any:
        js = super(_AnnoControlVis, self).to_json()
        js['visibility_label_text'] = self.visibility_label_text
        return js


class AnnoControlCheckbox (_AnnoControlVis):
    __control_type__ = 'checkbox'

    def __init__(self, identifier: str, label_text: str, visibility_label_text: Optional[str] = None):
        super(AnnoControlCheckbox, self).__init__(identifier, visibility_label_text=visibility_label_text)
        self.label_text = label_text

    def to_json(self) -> Any:
        js = super(AnnoControlCheckbox, self).to_json()
        js['label_text'] = self.label_text
        return js


class AnnoControlRadioButtons (_AnnoControlVis):
    __control_type__ = 'radio'

    def __init__(self, identifier: str, label_text: str, choices: List[Dict[str, Any]],
                 label_on_own_line: bool = False, visibility_label_text: Optional[str] = None):
        super(AnnoControlRadioButtons, self).__init__(identifier, visibility_label_text=visibility_label_text)
        self.label_text = label_text
        self.choices = choices
        self.label_on_own_line = label_on_own_line

    def to_json(self) -> Any:
        js = super(AnnoControlRadioButtons, self).to_json()
        js['label_text'] = self.label_text
        js['choices'] = self.choices
        js['label_on_own_line'] = self.label_on_own_line
        return js

    @classmethod
    def choice(cls, value, label_text, tooltip) -> Dict[str, Any]:
        return dict(value=value, label_text=label_text, tooltip=tooltip)


class AnnoControlPopupMenu (_AnnoControlVis):
    __control_type__ = 'popup_menu'

    def __init__(self, identifier: str, label_text: str, groups: List[Dict[str, Any]], visibility_label_text=None):
        super(AnnoControlPopupMenu, self).__init__(identifier, visibility_label_text=visibility_label_text)
        self.label_text = label_text
        self.groups = groups

    def to_json(self) -> Any:
        js = super(AnnoControlPopupMenu, self).to_json()
        js['label_text'] = self.label_text
        js['groups'] = self.groups
        return js

    @classmethod
    def group(cls, label_text, choices) -> Dict[str, Any]:
        return dict(label_text=label_text, choices=choices)

    @classmethod
    def choice(cls, value, label_text, tooltip) -> Dict[str, Any]:
        return dict(value=value, label_text=label_text, tooltip=tooltip)


class AnnoControlText (_AnnoControl):
    __control_type__ = 'text'

    def __init__(self, identifier: str, label_text: str, multiline: bool = False):
        super(AnnoControlText, self).__init__(identifier)
        self.label_text = label_text
        self.multiline = multiline

    def to_json(self) -> Any:
        js = super(AnnoControlText, self).to_json()
        js['label_text'] = self.label_text
        js['multiline'] = self.multiline
        return js


def _next_wrapped_array(xs: np.ndarray) -> np.ndarray:
    return np.append(xs[1:], xs[:1], axis=0)


def _prev_wrapped_array(xs: np.ndarray) -> np.ndarray:
    return np.append(xs[-1:], xs[:-1], axis=0)


def _simplify_contour(cs: np.ndarray) -> Optional[np.ndarray]:
    degenerate_verts = (cs == _next_wrapped_array(cs)).all(axis=1)
    while degenerate_verts.any():
        cs = cs[~degenerate_verts,:]
        degenerate_verts = (cs == _next_wrapped_array(cs)).all(axis=1)

    if cs.shape[0] > 0:
        # Degenerate edges
        edges = (_next_wrapped_array(cs) - cs)
        edges = edges / np.sqrt((edges**2).sum(axis=1))[:,None]
        degenerate_edges = (_prev_wrapped_array(edges) * edges).sum(axis=1) > (1.0 - 1.0e-6)
        cs = cs[~degenerate_edges,:]

        if cs.shape[0] > 0:
            return cs
    return None


_LABEL_CLASS_REGISTRY = {}


class ObjectTable:
    def __init__(self, id_prefix: Optional[str], objects: Optional[Sequence[Any]] = None):
        if id_prefix is None or id_prefix == '':
            id_prefix = str(uuid.uuid4())
        self._id_prefix = id_prefix
        self._object_id_to_obj = {}
        self._old_object_id_to_obj = {}
        self._next_object_idx = 1

        if objects is not None:
            # Register objects with object IDs
            for obj in objects:
                self.register(obj)

    def register(self, obj: Any) -> str:
        obj_id = obj.object_id

        if obj_id is None:
            obj_id = '{}__{}'.format(self._id_prefix, self._next_object_idx)
            self._next_object_idx += 1
            obj.object_id = obj_id
        elif isinstance(obj_id, int):
            self._next_object_idx = max(self._next_object_idx, obj_id + 1)
            obj_id = '{}__{}'.format(self._id_prefix, obj_id)
            obj.object_id = obj_id

        if obj_id in self._object_id_to_obj:
            if self._object_id_to_obj[obj_id] is not obj:
                raise ValueError('Duplicate object ID {}'.format(obj_id))
        else:
            self._object_id_to_obj[obj_id] = obj

        return obj_id

    def __getitem__(self, obj_id: Optional[str]) -> Optional[Any]:
        if obj_id is None:
            return None
        else:
            return self._object_id_to_obj[obj_id]

    def get(self, obj_id: Optional[str], default: Optional[Any] = None) -> Optional[Any]:
        if obj_id is None:
            return None
        else:
            return self._object_id_to_obj.get(obj_id, default)

    def _new_style_id(self, obj_id: Optional[Union[str, int]]) -> Optional[str]:
        """
        Convert object ID to new '<prefix_uuid>__<index>' style.
        Object IDs used to be integers; if `obj_id` is an int, then convert it to the new style.

        :param obj_id: [optional] object ID as either str (new style) or int (old style)
        :return: new style object ID as str or None if `obj_id` is None
        """
        if obj_id is None:
            return None
        elif isinstance(obj_id, int):
            return '{}__{}'.format(self._id_prefix, obj_id)
        else:
            return obj_id

    def __contains__(self, obj_id: Optional[str]) -> bool:
        return obj_id in self._object_id_to_obj


class LabelContext:
    def __init__(self, point_radius: float = 0.0):
        self.point_radius = point_radius


def label_cls(cls: Any) -> Any:
    """Label class decorator

    Registers a label class

    Exmaple:
    >>> @label_cls
    ... class SomeLabel (AbstractLabel):
    ...     __json_type_name__ = 'some'
    """
    json_label_type = cls.__json_type_name__
    _LABEL_CLASS_REGISTRY[json_label_type] = cls
    return cls


class AbstractLabel (object):
    __json_type_name__ = None

    def __init__(self, object_id: Optional[str] = None, classification: Optional[str] = None,
                 source: Optional[str] = None, anno_data: Optional[Dict[str, Any]] = None):
        """
        Constructor

        :param object_id: a unique integer object ID or None
        :param classification: a str giving the label's ground truth classification
        :param source: [optional] a str stating how the label was created
            (e.g. 'manual', 'auto:dextr', 'auto:maskrcnn', etc)
        :param anno_data: [optional] a dict mapping field names to values
        """
        self.object_id = object_id
        self.classification = classification
        self.source = source
        if anno_data is None:
            anno_data = {}
        self.anno_data = anno_data

    @property
    def dependencies(self) -> Sequence['AbstractLabel']:
        """Get a sequence of labels that `self` depends on

        :return: a sequence of label instances
        """
        return []

    def flatten(self) -> Generator['AbstractLabel', None, None]:
        """Get an iterator that yields a sequence of labels within this subtree, e.g. flattens groups

        :return: an iterator that yields label instances
        """
        yield self

    @staticmethod
    def flatten_json(label_json: Any) -> Generator[Any, None, None]:
        """Gets an iterator that yields a sequence of labels in JSON form within
        a subtree, e.g. flattens groups.

        Same idea as the `flatten` method, except this operates on labels stored as JSON rather than
        as objects.

        :return: an iterator that yields label instances in JSON form
        """
        label_type = label_json['label_type']
        cls = _LABEL_CLASS_REGISTRY.get(label_type)
        if cls is None:
            raise TypeError('Unknown label type {0}'.format(label_type))
        return cls._flatten_json_label(label_json)

    @classmethod
    def _flatten_json_label(cls, label_json: Any) -> Generator[Any, None, None]:
        """Helper method for `AbstractLabel.flatten_json`.

        :return: an iterator that yields label instances in JSON form
        """
        yield label_json

    def accumulate_label_class_histogram(self, histogram: MutableMapping[str, int]):
        """Accumulate a histogram of label classifications.
        Increments the count in `histogram` associated with the classification of this label.

        If `histogram` has no entry for the classification of this label, one will be created, e.g.:
        >>> label = AbstractLabel(classification='x')
        >>> hist = {}
        >>> label.accumulate_label_class_histogram(hist)
        >>> assert list(hist.keys()) == ['x']
        >>> assert hist['x'] == 1

        :param histogram: the histogram as a mapping, e.g. a dictionary
        """
        histogram[self.classification] = histogram.get(self.classification, 0) + 1

    @abstractmethod
    def bounding_box(self, ctx: Optional[LabelContext] = None) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        """Get an axis-aligned bounding box that surrounds self

        :param ctx: some context information that provides e.g. the size that a point label should occupy
        :return: bounding box as a tuple of `(min_xy, max_xy)` where `min_xy` and `max_xy` are `[x, y]` NumPy arrays
        """
        pass

    @abstractmethod
    def _warp(self, xform_fn: Callable[[np.ndarray], np.ndarray], object_table: ObjectTable) -> 'AbstractLabel':
        pass

    def warped(self, xform_fn: Callable[[np.ndarray], np.ndarray], object_table: Optional[ObjectTable] = None,
               id_prefix: Optional[str] = None):
        if object_table is None:
            object_table = ObjectTable(id_prefix)
        w = self._warp(xform_fn, object_table)
        object_table.register(w)
        return w

    @abstractmethod
    def _render_mask(self, img: Image, fill: bool, dx: float=0.0, dy: float=0.0,
                     ctx: Optional[LabelContext] = None):
        pass

    def render_mask(self, width: int, height: int, fill: bool, dx: float = 0.0, dy: float = 0.0,
                    ctx: Optional[LabelContext] = None):
        img = Image.new('L', (width, height), 0)
        self._render_mask(img, fill, dx, dy, ctx)
        return np.array(img)

    def to_json(self) -> Any:
        return dict(label_type=self.__json_type_name__,
                    object_id=self.object_id,
                    label_class=self.classification,
                    source=self.source,
                    anno_data=self.anno_data)

    @classmethod
    @abstractmethod
    def new_instance_from_json(cls, label_json: Any, object_table: ObjectTable) -> 'AbstractLabel':
        pass

    @staticmethod
    def from_json(label_json: Any, object_table: ObjectTable) -> 'AbstractLabel':
        label_type = label_json['label_type']
        cls = _LABEL_CLASS_REGISTRY.get(label_type)
        if cls is None:
            raise TypeError('Unknown label type {0}'.format(label_type))
        label = cls.new_instance_from_json(label_json, object_table)
        object_table.register(label)
        return label


@label_cls
class PointLabel (AbstractLabel):
    __json_type_name__ = 'point'

    def __init__(self, position_xy: np.ndarray, object_id: Optional[str] = None, classification: Optional[str] = None,
                 source: Optional[str] = None, anno_data: Optional[Dict[str, Any]] = None):
        """
        Constructor

        :param position_xy: position of point as a (2,) NumPy array providing the x and y co-ordinates
        :param object_id: a unique integer object ID or None
        :param classification: a str giving the label's ground truth classification
        :param source: [optional] a str stating how the label was created
            (e.g. 'manual', 'auto:dextr', 'auto:maskrcnn', etc)
        :param anno_data: [optional] a dict mapping field names to values
        """
        super(PointLabel, self).__init__(object_id, classification, source, anno_data)
        self.position_xy = np.array(position_xy).astype(float)

    def bounding_box(self, ctx: Optional[LabelContext] = None) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        point_radius = ctx.point_radius if ctx is not None else 0.0
        return self.position_xy - point_radius, self.position_xy + point_radius

    def _warp(self, xform_fn: Callable[[np.ndarray], np.ndarray], object_table: ObjectTable) -> AbstractLabel:
        warped_pos = xform_fn(self.position_xy[None, :])
        return PointLabel(warped_pos[0, :], self.object_id, self.classification, self.source, self.anno_data)

    def _render_mask(self, img: Image, fill: bool, dx: float=0.0, dy: float=0.0,
                     ctx: Optional[LabelContext] = None):
        point_radius = ctx.point_radius if ctx is not None else 0.0

        x = self.position_xy[0] + dx
        y = self.position_xy[1] + dy

        if point_radius == 0.0:
            ImageDraw.Draw(img).point((x, y), fill=1)
        else:
            ellipse = [(x-point_radius, y-point_radius),
                       (x+point_radius, y+point_radius)]
            if fill:
                ImageDraw.Draw(img).ellipse(ellipse, outline=1, fill=1)
            else:
                ImageDraw.Draw(img).ellipse(ellipse, outline=1, fill=0)

    def to_json(self) -> Any:
        js = super(PointLabel, self).to_json()
        js['position'] = dict(x=self.position_xy[0], y=self.position_xy[1])
        return js

    def __str__(self) -> str:
        return 'PointLabel(object_id={}, classification={}, position_xy={})'.format(
            self.object_id, self.classification, self.position_xy.tolist()
        )

    @classmethod
    def new_instance_from_json(cls, label_json: Any, object_table: ObjectTable) -> AbstractLabel:
        pos_xy = np.array([label_json['position']['x'], label_json['position']['y']])
        return PointLabel(pos_xy, label_json.get('object_id'),
                          classification=label_json['label_class'],
                          source=label_json.get('source'),
                          anno_data=label_json.get('anno_data'))


@label_cls
class PolygonLabel (AbstractLabel):
    __json_type_name__ = 'polygon'

    def __init__(self, regions: List[np.ndarray], object_id: Optional[str] = None,
                 classification: Optional[str] = None, source: Optional[str] = None,
                 anno_data: Optional[Dict[str, Any]] = None):
        """
        Constructor

        :param regions: list of regions where each region is an array of vertices as a (N,2) NumPy array providing the [x, y] co-ordinates
        :param object_id: a unique integer object ID or None
        :param classification: a str giving the label's ground truth classification
        :param source: [optional] a str stating how the label was created
            (e.g. 'manual', 'auto:dextr', 'auto:maskrcnn', etc)
        :param anno_data: [optional] a dict mapping field names to values
        """
        super(PolygonLabel, self).__init__(object_id, classification, source, anno_data)
        regions = [np.array(region).astype(float) for region in regions]
        self.regions = regions

    def bounding_box(self, ctx: Optional[LabelContext] = None) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        all_verts = np.concatenate(self.regions, axis=0)
        return all_verts.min(axis=0), all_verts.max(axis=0)

    def _warp(self, xform_fn: Callable[[np.ndarray], np.ndarray], object_table: ObjectTable) -> AbstractLabel:
        warped_regions = [xform_fn(region) for region in self.regions]
        return PolygonLabel(warped_regions, self.object_id, self.classification, self.source, self.anno_data)

    def _render_mask(self, img: Image, fill: bool, dx: float=0.0, dy: float=0.0,
                     ctx: Optional[LabelContext] = None):
        # Rendering helper function: create a binary mask for a given label

        # Polygonal label
        if fill:
            # Filled
            if len(self.regions) == 1:
                # Simplest case: 1 region
                region = self.regions[0]
                if len(region) >= 3:
                    vertices = region + np.array([[dx, dy]])
                    polygon = [tuple(v) for v in vertices]

                    ImageDraw.Draw(img).polygon(polygon, outline=1, fill=1)
            else:
                # Need to combine regions
                mask = np.zeros(img.size[::-1], dtype=bool)
                for region in self.regions:
                    if len(region) >= 3:
                        vertices = region + np.array([[dx, dy]])
                        polygon = [tuple(v) for v in vertices]

                        region_img = Image.new('L', img.size, 0)
                        ImageDraw.Draw(region_img).polygon(polygon, outline=1, fill=1)
                        region_img = np.array(region_img) > 0
                        mask = mask ^ region_img
                img_arr = np.array(img) | mask
                img.putdata(Image.fromarray(img_arr).getdata())
        else:
            # Outline only
            for region in self.regions:
                if len(region) >= 3:
                    vertices = region + np.array([[dx, dy]])
                    polygon = [tuple(v) for v in vertices]

                    ImageDraw.Draw(img).polygon(polygon, outline=1, fill=0)

    @staticmethod
    def regions_to_json(regions) -> Any:
        return [[dict(x=float(region[i,0]), y=float(region[i,1])) for i in range(len(region))]
                         for region in regions]

    def to_json(self) -> Any:
        js = super(PolygonLabel, self).to_json()
        js['regions'] = PolygonLabel.regions_to_json(self.regions)
        return js

    def __str__(self) -> str:
        return 'PolygonLabel(object_id={}, classification={}, regions={})'.format(
            self.object_id, self.classification, self.regions
        )

    @classmethod
    def new_instance_from_json(cls, label_json: Any, object_table: ObjectTable) -> AbstractLabel:
        if 'vertices' in label_json:
            regions_json = [label_json['vertices']]
        else:
            regions_json = label_json['regions']
        regions = [np.array([[v['x'], v['y']] for v in region_json]) for region_json in regions_json]
        return PolygonLabel(regions, label_json.get('object_id'),
                            classification=label_json['label_class'],
                            source=label_json.get('source'),
                            anno_data=label_json.get('anno_data'))

    @staticmethod
    def mask_image_to_regions(mask: np.ndarray) -> List[np.ndarray]:
        """
        Convert a mask label image to regions/contours that can be used as regions for a Polygonal label

        Uses Scikit-Image `find_contours`.

        :param mask: a mask image as  `(h,w)` numpy array (preferable of dtype bool) that identifies the pixels
            belonging to the object in question
        :return: regions as a list of NumPy arrays, where each array is (N, [x,y])
        """
        contours = []
        if mask.sum() > 0:
            mask_positions = np.argwhere(mask)
            (ystart, xstart), (ystop, xstop) = mask_positions.min(0), mask_positions.max(0) + 1

            if ystop >= ystart+1 and xstop >= xstart+1:
                mask_trim = mask[ystart:ystop, xstart:xstop]
                mask_trim = np.pad(mask_trim, [(1,1), (1,1)], mode='constant').astype(np.float32)
                cs = find_contours(mask_trim, 0.5)
                for contour in cs:
                    simp = _simplify_contour(contour + np.array((ystart, xstart)) - np.array([[1.0, 1.0]]))
                    if simp is not None:
                        contours.append(simp[:, ::-1])
        return contours

    @staticmethod
    def mask_image_to_regions_cv(mask: np.ndarray, sort_decreasing_area: bool = True) -> List[np.ndarray]:
        """
        Convert labels represented as a sequence of mask images to an `ImageLabels` instance.
        Mask to contour conversion performed using OpenCV `findContours`, finding external contours only.

        Raises RuntimeError is OpenCV is not available.

        :param mask: a mask image as  `(h,w)` numpy array (preferable of dtype bool) that identifies the pixels
            belonging to the object in question
        :param sort_decreasing_area: if True, regions are sorted in order of decreasing area
        :return: regions as a list of NumPy arrays, where each array is (N, [x,y])
        """
        if cv2 is None:
            raise RuntimeError('OpenCV is not available!')

        result = cv2.findContours((mask != 0).astype(np.uint8), cv2.RETR_LIST,
                                  cv2.CHAIN_APPROX_TC89_L1)
        if len(result) == 3:
            _, image_contours, _ = result
        else:
            image_contours, _ = result

        image_contours = [contour[:, 0, :] for contour in image_contours if len(contour) >= 3]

        if len(image_contours) > 0:
            # Compute area
            areas = ImageLabels._contour_areas(image_contours)

            if sort_decreasing_area:
                # Sort in order of decreasing area
                order = np.argsort(areas)[::-1]
                image_contours = [image_contours[i] for i in order]

        return image_contours


@label_cls
class BoxLabel (AbstractLabel):
    __json_type_name__ = 'box'

    def __init__(self, centre_xy: np.ndarray, size_xy: np.ndarray, object_id: Optional[str] = None,
                 classification: Optional[str] = None, source: Optional[str] = None,
                 anno_data: Optional[Dict[str, Any]] = None):
        """
        Constructor

        :param centre_xy: centre of box as a (2,) NumPy array providing the x and y co-ordinates
        :param size_xy: size of box as a (2,) NumPy array providing the x and y co-ordinates
        :param object_id: a unique integer object ID or None
        :param classification: a str giving the label's ground truth classification
        :param source: [optional] a str stating how the label was created
            (e.g. 'manual', 'auto:dextr', 'auto:maskrcnn', etc)
        :param anno_data: [optional] a dict mapping field names to values
        """
        super(BoxLabel, self).__init__(object_id, classification, source, anno_data)
        self.centre_xy = np.array(centre_xy).astype(float)
        self.size_xy = np.array(size_xy).astype(float)

    def bounding_box(self, ctx: Optional[LabelContext] = None) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        return self.centre_xy - self.size_xy * 0.5, self.centre_xy + self.size_xy * 0.5

    def _warp(self, xform_fn: Callable[[np.ndarray], np.ndarray], object_table: ObjectTable) -> AbstractLabel:
        corners = np.array([
            self.centre_xy + self.size_xy * -0.5,
            self.centre_xy + self.size_xy * np.array([0.5, -0.5]),
            self.centre_xy + self.size_xy * 0.5,
            self.centre_xy + self.size_xy * np.array([-0.5, 0.5]),
        ])
        xf_corners = xform_fn(corners)
        lower = xf_corners.min(axis=0)
        upper = xf_corners.max(axis=0)
        xf_centre = (lower + upper) * 0.5
        xf_size = upper - lower
        return BoxLabel(xf_centre, xf_size, self.object_id, self.classification, self.source, self.anno_data)

    def _render_mask(self, img: Image, fill: bool, dx: float=0.0, dy: float=0.0,
                     ctx: Optional[LabelContext] = None):
        # Rendering helper function: create a binary mask for a given label

        centre = self.centre_xy + np.array([dx, dy])
        lower = centre - self.size_xy * 0.5
        upper = centre + self.size_xy * 0.5

        if fill:
            ImageDraw.Draw(img).rectangle([tuple(lower), tuple(upper)], outline=1, fill=1)
        else:
            ImageDraw.Draw(img).rectangle([tuple(lower), tuple(upper)], outline=1, fill=0)

    def to_json(self) -> Any:
        js = super(BoxLabel, self).to_json()
        js['centre'] = dict(x=self.centre_xy[0], y=self.centre_xy[1])
        js['size'] = dict(x=self.size_xy[0], y=self.size_xy[1])
        return js

    def __str__(self) -> str:
        return 'BoxLabel(object_id={}, classification={}, centre_xy={}, size_xy={})'.format(
            self.object_id, self.classification, self.centre_xy.tolist(), self.size_xy.tolist()
        )

    @classmethod
    def new_instance_from_json(cls, label_json: Any, object_table: ObjectTable) -> AbstractLabel:
        centre = np.array([label_json['centre']['x'], label_json['centre']['y']])
        size = np.array([label_json['size']['x'], label_json['size']['y']])
        return BoxLabel(centre, size, label_json.get('object_id'),
                        classification=label_json['label_class'],
                        source=label_json.get('source'),
                        anno_data=label_json.get('anno_data'))


@label_cls
class OrientedEllipseLabel (AbstractLabel):
    __json_type_name__ = 'oriented_ellipse'

    def __init__(self, centre_xy: np.ndarray, radius1: float, radius2: float, orientation_rad: float,
                 object_id: Optional[str] = None,
                 classification: Optional[str] = None, source: Optional[str] = None,
                 anno_data: Optional[Dict[str, Any]] = None):
        """
        Constructor

        :param centre_xy: centre of box as a (2,) NumPy array providing the x and y co-ordinates
        :param radius1: radius in orientation axis (X-axis rotated clockwise by `orientation_rad`)
        :param radius2: radius in axis perpendicular to orientation axis
        :param orientation_rad: orientation/rotation measured in radians clockwise from positive x-axis,
            so orientation rotates the positive x-axis (right) to the positive y-axis (down)
        :param object_id: a unique integer object ID or None
        :param classification: a str giving the label's ground truth classification
        :param source: [optional] a str stating how the label was created
            (e.g. 'manual', 'auto:dextr', 'auto:maskrcnn', etc)
        :param anno_data: [optional] a dict mapping field names to values
        """
        super(OrientedEllipseLabel, self).__init__(object_id, classification, source, anno_data)
        self.centre_xy = np.array(centre_xy).astype(float)
        self.radius1 = radius1
        self.radius2 = radius2
        self.orientation_rad = orientation_rad

    def bounding_box(self, ctx: Optional[LabelContext] = None) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        # The solution to this problem was obtained from here:
        # https://stackoverflow.com/questions/87734/how-do-you-calculate-the-axis-aligned-bounding-box-of-an-ellipse
        # https://gist.github.com/smidm/b398312a13f60c24449a2c7533877dc0
        # Note that orientation rotates clockwise from positive X-axis (right) to the positive Y-axis (down)
        # If the orientation is 0, then radius1 and radius2 correspond to the radii in the X and Y axes respectively
        tan_orient = math.tan(self.orientation_rad)
        s0 = math.atan(-self.radius2 * tan_orient / self.radius1)
        s1 = s0 + math.pi
        if tan_orient != 0.0:
            t0 = math.atan((self.radius2 / tan_orient) / self.radius1)
        else:
            t0 = math.pi * 0.5
        t1 = t0 + math.pi
        max_x, min_x = [self.centre_xy[0] + self.radius1 * math.cos(s) * math.cos(self.orientation_rad) -
                        self.radius2 * math.sin(s) * math.sin(self.orientation_rad) for s in [s0, s1]]
        max_y, min_y = [self.centre_xy[1] + self.radius2 * math.sin(s) * math.cos(self.orientation_rad) +
                        self.radius1 * math.cos(s) * math.sin(self.orientation_rad) for s in [t0, t1]]
        return np.array([min_x, min_y]), np.array([max_x, max_y])

    def _warp(self, xform_fn: Callable[[np.ndarray], np.ndarray], object_table: ObjectTable) -> AbstractLabel:
        u_xy = np.array([math.cos(self.orientation_rad), math.sin(self.orientation_rad)])
        v_xy = np.array([-math.sin(self.orientation_rad), math.cos(self.orientation_rad)])
        u_points_xy = self.centre_xy[None, :] + u_xy[None, :] * self.radius1 * np.array([-1.0, 1.0])[:, None]
        v_point_xy = self.centre_xy + v_xy * self.radius2
        uv = np.append(u_points_xy, v_point_xy[None, :], axis=0)
        uv = xform_fn(uv)
        return OrientedEllipseLabel.new_instance_from_uv_points(
            uv[0:2], uv[2], self.object_id, self.classification, self.source, self.anno_data)

    def _render_mask(self, img: Image, fill: bool, dx: float=0.0, dy: float=0.0,
                     ctx: Optional[LabelContext] = None):
        # We will draw it as a polygon. We want the vertices to be space at most 1 pixel apart, so take
        # the maximum radius and compute the circumference of a circle of that radius and round
        n_thetas = int(round(2.0 * math.pi * max(self.radius1, self.radius2)))
        thetas = np.linspace(0.0, 2.0 * math.pi, n_thetas + 1)[:-1]
        vx = self.centre_xy[0] + dx + self.radius1 * np.cos(thetas) * math.cos(self.orientation_rad) - \
                                      self.radius2 * np.sin(thetas) * math.sin(self.orientation_rad)
        vy = self.centre_xy[1] + dy + self.radius2 * np.sin(thetas) * math.cos(self.orientation_rad) + \
                                      self.radius1 * np.cos(thetas) * math.sin(self.orientation_rad)
        polygon = list(zip(vx, vy))
        ImageDraw.Draw(img).polygon(polygon, outline=1, fill=(1 if fill else 0))

    def to_json(self) -> Any:
        js = super(OrientedEllipseLabel, self).to_json()
        js['centre'] = dict(x=self.centre_xy[0], y=self.centre_xy[1])
        js['radius1'] = self.radius1
        js['radius2'] = self.radius2
        js['orientation_radians'] = self.orientation_rad
        return js

    def __str__(self) -> str:
        return 'OrientedEllipseLabel(object_id={}, classification={}, centre_xy={}, radius1={}, radius2={}, ' \
               'orientation={})'.format(self.object_id, self.classification, self.centre_xy.tolist(), self.radius1,
                                        self.radius2, self.orientation_rad)

    @classmethod
    def uv_points_to_params(cls, u_points_xy: np.ndarray, v_point_xy: np.ndarray) -> \
                Tuple[np.ndarray, float, float, float]:
        """Compute oriented ellipse parameters from three points; two end points that defines the U-axis that
        provides the orientation and whose distance apart is twice `radius1` and a 3rd point whose distance
        from the U-axis is the second radius.

        This is how the client-side Javascript tool constructs the ellipse from mouse clicks.

        :param u_points_xy: end points defining the U-axis as a `(2, [x, y])` NumPy array
        :param v_point_xy: 3rd point as a `[x, y]` NumPy array
        :return: (centre_xy, radius1, radius2, orientation) where `centre_xy` is a `[x, y]` NumPy array,
            `radius1` and `radius2` are the radii in the U-axis and perpendicular V-axis and `orientation`
            is the orientation of the U-axis measured in radians, clockwise from the positive X-axis
        """
        u = u_points_xy[1] - u_points_xy[0]
        centre_xy = (u_points_xy[0] + u_points_xy[1]) * 0.5
        radius1_x2 = np.sqrt(u @ u)
        radius1 = radius1_x2 * 0.5
        orientation = math.atan2(float(u[1]), float(u[0]))
        u_nrm = u / radius1_x2
        n = np.array([u_nrm[1], -u_nrm[0]])
        d_origin_u = n @ u_points_xy[0]
        d_origin_v = n @ v_point_xy
        radius2 = abs(d_origin_v - d_origin_u)
        return centre_xy, radius1, radius2, orientation

    @classmethod
    def new_instance_from_uv_points(cls, u_points_xy: np.ndarray, v_point_xy: np.ndarray,
                                    object_id: Optional[str] = None,
                                    classification: Optional[str] = None, source: Optional[str] = None,
                                    anno_data: Optional[Dict[str, Any]] = None) -> 'OrientedEllipseLabel':
        centre_xy, radius1, radius2, orientation = cls.uv_points_to_params(u_points_xy, v_point_xy)
        return OrientedEllipseLabel(centre_xy, radius1, radius2, orientation, object_id, classification, source,
                                    anno_data)

    @classmethod
    def new_instance_from_json(cls, label_json: Any, object_table: ObjectTable) -> AbstractLabel:
        centre = np.array([label_json['centre']['x'], label_json['centre']['y']])
        return OrientedEllipseLabel(
            centre, label_json['radius1'], label_json['radius2'], label_json['orientation_radians'],
            label_json.get('object_id'), classification=label_json['label_class'], source=label_json.get('source'),
            anno_data=label_json.get('anno_data'))


@label_cls
class CompositeLabel (AbstractLabel):
    __json_type_name__ = 'composite'

    def __init__(self, components, object_id=None, classification=None, source=None, anno_data=None):
        """
        Constructor

        :param components: a list of label objects that are members of the composite label
        :param object_id: a unique integer object ID or None
        :param classification: a str giving the label's ground truth classification
        :param source: [optional] a str stating how the label was created
            (e.g. 'manual', 'auto:dextr', 'auto:maskrcnn', etc)
        :param anno_data: [optional] a dict mapping field names to values
        """
        super(CompositeLabel, self).__init__(object_id, classification, source, anno_data)
        self.components = components

    @property
    def dependencies(self) -> Sequence[AbstractLabel]:
        return self.components

    def bounding_box(self, ctx: Optional[LabelContext] = None) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        return None, None

    def _warp(self, xform_fn: Callable[[np.ndarray], np.ndarray], object_table: ObjectTable) -> AbstractLabel:
        warped_components = []
        for comp in self.components:
            if comp.object_id in object_table:
                warped_comp = object_table[comp.object_id]
            else:
                warped_comp = comp.warped(xform_fn, object_table)
            warped_components.append(warped_comp)
        return CompositeLabel(warped_components, self.object_id, self.classification, self.source, self.anno_data)

    def _render_mask(self, img: Image, fill: bool, dx: float=0.0, dy: float=0.0,
                     ctx: Optional[LabelContext] = None):
        return None

    def to_json(self) -> Any:
        js = super(CompositeLabel, self).to_json()
        js['components'] = [component.object_id for component in self.components]
        return js

    def __str__(self) -> str:
        return 'CompositeLabel(object_id={}, classification={}, ids(components)={}'.format(
            self.object_id, self.classification, [c.object_id for c in self.components]
        )

    @classmethod
    def new_instance_from_json(cls, label_json: Any, object_table: ObjectTable) -> AbstractLabel:
        component_ids = [object_table._new_style_id(obj_id) for obj_id in label_json['components']]
        components = [object_table.get(obj_id) for obj_id in component_ids]
        components = [comp for comp in components if comp is not None]
        return CompositeLabel(components, label_json.get('object_id'),
                              classification=label_json['label_class'],
                              source=label_json.get('source'),
                              anno_data=label_json.get('anno_data'))


@label_cls
class GroupLabel (AbstractLabel):
    __json_type_name__ = 'group'

    def __init__(self, component_labels, object_id=None, classification=None, source=None, anno_data=None):
        """
        Constructor

        :param component_labels: a list of label objects that are members of the group label
        :param object_id: a unique integer object ID or None
        :param classification: a str giving the label's ground truth classification
        :param source: [optional] a str stating how the label was created
            (e.g. 'manual', 'auto:dextr', 'auto:maskrcnn', etc)
        :param anno_data: [optional] a dict mapping field names to values
        """
        super(GroupLabel, self).__init__(object_id, classification, source, anno_data)
        self.component_labels = component_labels

    def __len__(self):
        return len(self.component_labels)

    def __getitem__(self, item):
        return self.component_labels[item]

    def flatten(self) -> Generator[AbstractLabel, None, None]:
        for comp in self.component_labels:
            for f in comp.flatten():
                yield f
        yield self

    @classmethod
    def _flatten_json_label(cls, label_json: Any) -> Generator[Any, None, None]:
        """Helper method for `AbstractLabel.flatten_json`.

        :return: an iterator that yields label instances in JSON form
        """
        for comp_json in label_json['component_models']:
            for f_json in AbstractLabel.flatten_json(comp_json):
                yield f_json
        yield label_json

    def bounding_box(self, ctx: Optional[LabelContext] = None) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        lowers, uppers = list(zip(*[comp.bounding_box(ctx) for comp in self.component_labels]))
        lowers = [x for x in lowers if x is not None]
        uppers = [x for x in uppers if x is not None]
        if len(lowers) > 0 and len(uppers) > 0:
            return np.array(lowers).min(axis=0), np.array(uppers).max(axis=0)
        else:
            return None, None

    def _warp(self, xform_fn: Callable[[np.ndarray], np.ndarray], object_table: ObjectTable) -> AbstractLabel:
        comps = [comp.warped(xform_fn, object_table) for comp in self.component_labels]
        return GroupLabel(comps, self.object_id, self.classification, self.source, self.anno_data)

    def _render_mask(self, img: Image, fill: bool, dx: float=0.0, dy: float=0.0,
                     ctx: Optional[LabelContext] = None):
        for label in self.component_labels:
            label._render_mask(img, fill, dx, dy, ctx)

    def to_json(self) -> Any:
        js = super(GroupLabel, self).to_json()
        js['component_models'] = [component.to_json() for component in self.component_labels]
        return js

    def __str__(self) -> str:
        return 'GroupLabel(object_id={}, classification={}, component_labels={}'.format(
            self.object_id, self.classification, self.component_labels
        )

    @classmethod
    def new_instance_from_json(cls, label_json: Any, object_table: ObjectTable) -> AbstractLabel:
        components = [AbstractLabel.from_json(comp, object_table)
                      for comp in label_json['component_models']]
        return GroupLabel(components, label_json.get('object_id'),
                          classification=label_json['label_class'],
                          source=label_json.get('source'),
                          anno_data=label_json.get('anno_data'))


_ClassIndexMappingFunction = Callable[[str], Optional[int]]
_ClassIndexMappingMap = Mapping[str, int]
_ClassIndexMappingCls = Union[str, LabelClass, None]
_ClassIndexMappingListEntry = Union[str, LabelClass, None, Sequence[_ClassIndexMappingCls]]
_ClassIndexMappingList = Sequence[_ClassIndexMappingListEntry]

ClassIndexMapping = Union[_ClassIndexMappingFunction,
                          _ClassIndexMappingMap,
                          _ClassIndexMappingList]

class ImageLabels:
    """
    Represents labels in vector format, stored in JSON form. Has methods for
    manipulating and rendering them.

    """
    def __init__(self, labels: List[AbstractLabel], obj_table: Optional[ObjectTable] = None,
                 id_prefix: Optional[str] = None):
        self.labels = labels
        if obj_table is None:
            obj_table = ObjectTable(id_prefix, list(self.flatten()))
        self._obj_table = obj_table

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, item: Union[int, str, slice, Sequence[Union[str, int]]]) -> \
                    Union[AbstractLabel, 'ImageLabels']:
        if isinstance(item, int):
            return self.labels[item]
        elif isinstance(item, str):
            return self._obj_table[item]
        elif isinstance(item, slice):
            return self.retain(item)
        elif isinstance(item, Sequence):
            return self.retain(item)
        else:
            raise TypeError('item should be an int, a str, a slice or a sequence, not a {}'.format(type(item)))

    def flatten(self) -> Generator[AbstractLabel, None, None]:
        for lab in self.labels:
            for f in lab.flatten():
                yield f

    def label_class_histogram(self) -> Mapping[str, int]:
        histogram = {}
        for lab in self.labels:
            lab.accumulate_label_class_histogram(histogram)
        return histogram

    def retain(self, items: Union[slice, Sequence[Union[str, int]]], id_prefix: Optional[str] = None) -> 'ImageLabels':
        """
        Create a clone of the labels listed in `items`

        :param items: Either a slice, or a list of indices/object IDs that identify the labels to be kept
        :param id_prefix: the object ID prefix that will be to create object IDs for labels added to
            the returned `ImageLabels` instance
        :return: `ImageLabels` instance
        """
        if isinstance(items, slice):
            retained_labels = copy.deepcopy(self.labels[items])
        else:
            retained_labels = []
            for item in items:
                if isinstance(item, str):
                    retained_labels.append(copy.deepcopy(self._obj_table[item]))
                else:
                    retained_labels.append(copy.deepcopy(self.labels[item]))

        if id_prefix is None:
            id_prefix = str(uuid.uuid4())
        obj_table = ObjectTable(id_prefix=id_prefix)

        return ImageLabels(retained_labels, obj_table=obj_table)

    def replace_label_classes(self, replacements: Mapping[Optional[str], Optional[str]]):
        """Replace all uses of a source classification with a target classification

        For each source-target pair in the mapping `replacements`, any labels whose classification is
        source will have their classification set to target.

        :param replacements: a dictionary mapping source to target class.
        """
        for label in self.flatten():
            if label.classification in replacements:
                label.classification = replacements[label.classification]

    @staticmethod
    def replace_label_classes_json(labels_json, replacements: Mapping[Optional[str], Optional[str]]):
        """Replace all uses of a source classification with a target classification in a list of labels in JSON form

        For each source-target pair in the mapping `replacements`, any labels whose classification is
        source will have their classification set to target.

        :param labels_json: labels in JSON form
        :param replacements: a dictionary mapping source to target class.
        """
        for label_js in labels_json:
            for f_json in AbstractLabel.flatten_json(label_js):
                if f_json['label_class'] in replacements:
                    f_json['label_class'] = replacements[f_json['label_class']]

    def warp(self, xform_fn: Callable[[np.ndarray], np.ndarray]) -> 'ImageLabels':
        """
        Warp the labels given a warping function

        :param xform_fn: a transformation function of the form `f(vertices) -> warped_vertices`, where `vertices` and
        `warped_vertices` are both Numpy arrays of shape `(N, [x, y])` where `N` is the number of vertices and the
        co-ordinates are `x,y` pairs. The transformations defined in `skimage.transform`, e.g. `AffineTransform` can
        be used here.
        :return: an `ImageLabels` instance that contains the warped labels
        """
        warped_obj_table = ObjectTable(id_prefix=str(uuid.uuid4()))
        warped_labels = [lab.warped(xform_fn, warped_obj_table) for lab in self.labels]
        return ImageLabels(warped_labels, obj_table=warped_obj_table)

    def _label_class_list_to_mapping_fn(self, label_classes: ClassIndexMapping, start_at: int = 0) -> \
            Tuple[Callable[[str], Optional[int]], int]:
        """
        Coerce `label_classes` to a tuple of a mapping function that maps label class name (string) to class index
        (or None to skip) and the number of classes.

        If `label_classes` is callable then use it as is. We do however invoke `label_classes` for the class
            of every label and take the maximum class index returned in order to determine the number of classes.

        If `label_classes` is a mapping, return its `get` method, along with the number of classes.

        If `label_classes` is a list, then each entry should be a class (class name as a str or
            a `LabelClass` instance) or list of classes (str/LabelClass).
            The class or classes in each entry is/are mapped to the position of the entry, with the offset
            `start_at` applied.
            For example `['a', ['b', 'c'], 'd']` will map the class 'a' to 0, 'b' and 'c' to 1 and 'd' to 2.

        :param label_classes: the mapping as a function, dictionary or list.
        :param start_at: class index offset used if `label_classes` is a list, ignored otherwise
        :return: `(cls_name_to_index_fn, n_classes)` where `cls_name_to_index_fn` is a function of the form
            `f(class_name) -> index` and `n_classes` is the number of classes
        """
        if isinstance(label_classes, Mapping):
            # Dict, compute number of classes and return along with `get` method
            n_classes = max(label_classes.values()) + 1
            return label_classes.get, n_classes
        elif callable(label_classes):
            # Function: determine number of classes and return
            n_classes = 0
            for label in self.labels:
                cls_i = label_classes(label.classification)
                if cls_i is not None:
                    n_classes = max(n_classes, cls_i + 1)
            return label_classes, n_classes
        elif isinstance(label_classes, Sequence):
            # List
            cls_to_index = {}
            for i, cls in enumerate(label_classes):
                if isinstance(cls, LabelClass):
                    cls_to_index[cls.name] = i + start_at
                elif isinstance(cls, str)  or  cls is None:
                    cls_to_index[cls] = i + start_at
                elif isinstance(cls, list)  or  isinstance(cls, tuple):
                    for c in cls:
                        if isinstance(c, LabelClass):
                            cls_to_index[c.name] = i + start_at
                        elif isinstance(c, str)  or  c is None:
                            cls_to_index[c] = i + start_at
                        else:
                            raise TypeError('Item {0} in label_classes is a list that contains an item that is not a '
                                            'LabelClass or a string but a {1}'.format(i, type(c).__name__))
                else:
                    raise TypeError('Item {0} in label_classes is not a LabelClass, string or list, '
                                    'but a {1}'.format(i, type(cls).__name__))
            n_classes = len(label_classes)
            return (lambda cls_name: cls_to_index.get(cls_name, None)), n_classes
        else:
            raise TypeError('label_classes must be a dict or a sequence. The sequence can contain LabelClass '
                            'instances, strings or nested sequences of the former')

    def render_label_classes(self, label_classes: ClassIndexMapping, image_shape: Tuple[int, int],
                             multichannel_mask: bool = False, fill: bool = True, ctx: Optional[LabelContext] = None):
        """Render label classes to a create a label class image suitable for use as a
        semantic segmentation ground truth image.

        If `multichannel_mask` is False, then a label class image is generated. A label class image uses integer
        pixel values to indicate the class of the label the pixel belongs to. They cannot represent overlapping labels;
        in such cases labels that come later in the list will appear 'in front' of prior labels, overwriting them.
        A label class image is an array of shape `(height, width)` with integer dtype. A value of 0 represents
        background/empty. Each label is drawn into the image, with the value determined by its class index. Its
        class index is determined by the mapping provided as the `label_classes` parameter.

        If `multichannel_mask` is True, then a multi-channel mask is generated. A multi-channel mask
        is an array of shape `(height, width, n_classes)` with boolean dtype. Each label is drawn into the channel
        with the value `True` where the channel index is determined by its class index. Once again, the class
        index for a label is determined using the `label_classes` mapping.

        The label class mapping (`label_classes` parameter) is provided as either a callable/function,
        dictionary or sequence of classes:
            Function of the form `fn(class_name) -> class_index`: the function returns the class index given
                the label class name. If it returns `None` the label will be skipped. Note that when
                rendering a label image (`multichannel_mask` is `False`) the function should *not* return 0
                as that is the index given to the background.
            Dictionary: maps class name to class index. Uses the same rules when using a function, as above.
                We effectively use the `get` method of the dictionary as the mapping function.
            Sequence: each entry in the sequence should be a class (class name as a str or a `LabelClass` instance)
                or list of classes (str/LabelClass). The class or classes in each entry is/are mapped to the
                position of the entry, where the position starts at 1 for label images or 0 for multi-channel
                masks. For example `['a', ['b', 'c'], 'd']` will map the class 'a' to class index 1 for
                a label class image or channel 0 for multi-channel mask, 'b' and 'c' to class index 2/channel 1
                and 'd' to class index 3/channel 2.

        :param label_classes: label class mapping as described above
        :param image_shape: `(height, width)` tuple specifying the shape of the image to be returned
        :param multichannel_mask: If `False`, return a label image (`(height, width)` shaped array of dtype=int)
            If `True` return a multi-channel mask (`(height, width, n_classes)` array of dtype=bool)
        :param fill: if True, labels will be filled, otherwise their outlines will be drawn
        :param ctx: [optional] a `LabelContext` instance that provides parameters that control the rendering
            of some labels (e.g. point labels)
        :return: label image as (H,W) array with dtype=int or multi-channel mask as (H,W,n_classes) array
            with dtype=bool
        """
        cls_to_index_fn, n_classes = self._label_class_list_to_mapping_fn(
            label_classes, 0 if multichannel_mask else 1)

        height, width = image_shape

        if multichannel_mask:
            label_image = np.zeros((height, width, n_classes), dtype=bool)
        else:
            label_image = np.zeros((height, width), dtype=int)

        for label in self.labels:
            label_cls_n = cls_to_index_fn(label.classification,)
            if label_cls_n is not None:
                mask = label.render_mask(width, height, fill, ctx=ctx)
                if mask is not None:
                    mask = mask >= 0.5
                    if multichannel_mask:
                        label_image[:,:,label_cls_n] |= mask
                    else:
                        label_image[mask] = label_cls_n

        return label_image

    def render_label_instances(self, label_classes: Optional[ClassIndexMapping], image_shape: Tuple[int, int],
                               multichannel_mask: bool = False, fill: bool = True,
                               return_object_ids: bool = False, ctx: Optional[LabelContext] = None):
        """Render a label instance image suitable for use as an instance segmentation ground truth image.
        Can render either a label image or a multi-channel mask.

        If `multichannel_mask` is False, then a label image is generated. A label image uses integer pixel values
        to indicate which label the pixel belongs to. They cannot represent overlapping labels; in such cases labels
        that come later in the list will appear 'in front' of prior labels. A label image is an array of shape
        `(height, width)` with integer dtype. A value of 0 represents background/empty. Each label is drawn
        into the image, with an instance index starting at 1, incrementing with each label that is drawn.
        Labels skipped as a result of the `label_classes` mapping indicating as such will not cause the instance
        index to increment).

        A multi-channel mask image has a boolean dtype and uses one channel or mask for each label/instance, where
        a pixel value of `True` indicates that the corresponding label covers that pixel. Overlapping labels are
        faithfully represented. A multi-channel mask is an array of shape `(height, width, n_instances)` with
        boolean dtype. Each label/instance is drawn into a `(height, width)` shaped mask image that is added to
        the stack of masks/channels. Labels that are skipped as indicated by the mapping in `label_classes`
        do *not* result add anything to the result.

        In addition to the label image or multi-channel mask this method also returns `label_index_to_cls`; a
        label index to class mapping that gives the class index of each label that was rendered.
        If `self` consists of one label and that label class is mapped to a class index of 5,
        if `multichannel_mask` is `False` this method will render a label image with a value of 1 in the
        pixels covered by the label and `label_index_to_cls` will have the values `[0, 5]`, with the first entry
        corresponding to the background. If `multichannel_mask` is `True` the resulting multi-channel mask
        will have one channel that will be `True` for pixels covered by the label and `label_index_to_cls`
        will have the value `[5]`.

        If `return_object_ids` is `True` then the unique object IDs (a UUID strong followed by underscores and
        an index) of the rendered labels will also be returned. The list of object IDs will follow the same form
        as `label_index_to_cls`; if `multichannel_mask` is `False` then the first entry that corresponds to the
        background will have an object ID of `None`, otherwise if `True` then the background will not have
        an object ID.

        The class indices are determined by thee label class mapping (`label_classes` parameter) that is provided
        as either a callable/function, dictionary, sequence of classes or None:
            Function of the form `fn(class_name) -> class_index`: the function returns the class index given
                the label class name. If it returns `None` the label will be skipped. Returning 0
                does not pose the same problems as with the `render_label_classes` method as the background
                class 0 only appears in the first entry of `label_index_to_cls`.
            Dictionary: maps class name to class index. Uses the same rules when using a function, as above.
                We effectively use the `get` method of the dictionary as the mapping function.
            Sequence: each entry in the sequence should be a class (class name as a str or a `LabelClass` instance)
                or list of classes (str/LabelClass). The class or classes in each entry is/are mapped to the
                position of the entry, where the position starts at 1. For example `['a', ['b', 'c'], 'd']` will
                map the class 'a' to 1, 'b' and 'c' to 2 and 'd' to 3.
            None: all labels will be rendered and assigned a class index of 1

        :param label_classes: label class mapping as described above
        :param image_shape: `(height, width)` tuple specifying the shape of the image to be returned
        :param multichannel_mask: If `False`, return a label image (`(height, width)` shaped array of dtype=int)
            If `True` return a multi-channel mask (`(height, width, n_classes)` array of dtype=bool)
        :param fill: if True, labels will be filled, otherwise their outlines will be drawn
        :param return_object_ids: if True, the returned tuple will contain a list that gives the object ID of
            each instance
        :param ctx: [optional] a `LabelContext` instance that provides parameters that control the rendering
            of some labels (e.g. point labels)
        :return: tuple of (label_image, label_index_to_cls) or (label_image, label_index_to_cls, object_ids) where:
            label_image is a (H,W) or (H,W,N) array with dtype=int
            label_index_to_cls is a 1D array that gives the class index of each labels. If `multichannel_mask` is
                False, the first entry at index 0 will have a value of 0 as it is the background label.
            object_ids: a list containing the object ID for each label/instance (only present if return_object_ids
                is True). If `multichannel_mask` is False, the first entry at index 0 will have a value of None as it
                is the background label.
        """
        if label_classes is not None:
            cls_to_index_fn, _ = self._label_class_list_to_mapping_fn(label_classes, 1)
        else:
            cls_to_index_fn = None

        height, width = image_shape

        if multichannel_mask:
            label_image = None
            label_image_stack = []
        else:
            label_image = np.zeros((height, width), dtype=int)
            label_image_stack = None

        if multichannel_mask:
            label_i = 0
            label_index_to_cls = []
            object_ids = []
        else:
            label_i = 1
            label_index_to_cls = [0]
            object_ids = [None]


        for label in self.labels:
            if cls_to_index_fn is not None:
                label_cls = cls_to_index_fn(label.classification)
            else:
                label_cls = 1
            if label_cls is not None:
                mask = label.render_mask(width, height, fill, ctx=ctx)
                if mask is not None:
                    mask = mask >= 0.5
                    if multichannel_mask:
                        label_image_stack.append(mask)
                    else:
                        label_image[mask] = label_i
                    label_index_to_cls.append(label_cls)
                    object_ids.append(label.object_id)
                    label_i += 1

        if multichannel_mask:
            if len(label_image_stack) > 0:
                label_image = np.stack(label_image_stack, axis=2)
            else:
                label_image = np.zeros((height, width, 0), dtype=int)

        if return_object_ids:
            return label_image, np.array(label_index_to_cls), object_ids
        else:
            return label_image, np.array(label_index_to_cls)

    def extract_label_images(self, image_2d: np.ndarray, label_class_set: Optional[Container[str]]=None,
                             ctx: Optional[LabelContext]=None):
        """Extract an image of each labelled entity from a given image.
        The resulting image is the original image masked with an alpha channel that results from rendering the label

        :param image_2d: the image from which to extract images of labelled objects
        :param label_class_set: a set or sequence of classes whose labels should be rendered, or None for all labels
        :param ctx: [optional] a `LabelContext` instance that provides parameters
        :return: a list of (H,W,C) image arrays
        """
        image_shape = image_2d.shape[:2]

        label_images = []

        for label in self.labels:
            if label_class_set is None or label.classification in label_class_set:
                bounds = label.bounding_box(ctx=ctx)

                if bounds[0] is not None and bounds[1] is not None:
                    lx = int(math.floor(bounds[0][0]))
                    ly = int(math.floor(bounds[0][1]))
                    ux = int(math.ceil(bounds[1][0]))
                    uy = int(math.ceil(bounds[1][1]))

                    # Given that the images and labels may have been warped by a transformation,
                    # there is no guarantee that they lie within the bounds of the image
                    lx = max(min(lx, image_shape[1]), 0)
                    ux = max(min(ux, image_shape[1]), 0)
                    ly = max(min(ly, image_shape[0]), 0)
                    uy = max(min(uy, image_shape[0]), 0)

                    w = ux - lx
                    h = uy - ly

                    if w > 0 and h > 0:
                        mask = label.render_mask(w, h, fill=True, dx=float(-lx), dy=float(-ly), ctx=ctx)
                        if mask is not None and (mask > 0).any():
                            img_box = image_2d[ly:uy, lx:ux]
                            if len(img_box.shape) == 2:
                                # Convert greyscale image to RGB:
                                img_box = gray2rgb(img_box)
                            # Append the mask as an alpha channel
                            object_img = np.append(img_box, mask[:, :, None] * 255, axis=2)

                            label_images.append(object_img)

        return label_images

    def to_json(self) -> Any:
        return [lab.to_json() for lab in self.labels]

    def replace_json(self, existing_json: Any) -> Any:
        if isinstance(existing_json, dict):
            new_dict = {}
            new_dict.update(existing_json)
            new_dict['labels'] = self.to_json()
            return new_dict
        elif isinstance(existing_json, list):
            return self.to_json()
        else:
            raise ValueError('existing_json should be a list or a dict')

    @deprecated('This methods will be removed in the future')
    def wrapped_json(self, image_filename: str, completed_tasks: List[str]) -> Any:
        return {'image_filename': image_filename,
                'completed_tasks': completed_tasks,
                'labels': self.to_json()}

    @classmethod
    def merge(cls, *image_labels: 'ImageLabels') -> 'ImageLabels':
        """
        Merge multiple `ImageLabel` label collections.

        :param image_labels: `ImageLabel` instances to merge
        :return: `ImageLabels` instance
        """
        obj_table = ObjectTable(id_prefix=str(uuid.uuid4()))
        merged_labels = []
        for il in image_labels:
            merged_labels.extend(copy.deepcopy(il.labels))
        used_ids = set()
        for label in merged_labels:
            for f_label in label.flatten():
                if f_label.object_id in used_ids:
                    f_label.object_id = None
                else:
                    used_ids.add(f_label.object_id)
        return ImageLabels(merged_labels, obj_table=obj_table)

    @staticmethod
    def from_json(label_data_js: Any, id_prefix: Optional[str] = None) -> 'ImageLabels':
        """
        Load from labels in JSON format

        :param label_data_js: either a list of labels in JSON format or a dict that maps the key `'labels'` to a list
        of labels in JSON form. The dict format will match the format stored in JSON label files.

        :return: an `ImageLabels` instance
        """
        if isinstance(label_data_js, dict):
            if 'labels' not in label_data_js:
                raise ValueError('label_js should be a list or a dict containing a \'labels\' key')
            labels = label_data_js['labels']
            if not isinstance(labels, list):
                raise TypeError('labels[\'labels\'] should be a list')
        elif isinstance(label_data_js, list):
            labels = label_data_js
        else:
            raise ValueError('label_data_js should be a list or a dict containing a \'labels\' key, it is a {}'.format(
                type(label_data_js)
            ))

        if id_prefix is None:
            id_prefix = str(uuid.uuid4())
        obj_table = ObjectTable(id_prefix=id_prefix)
        labs = [AbstractLabel.from_json(label, obj_table) for label in labels]
        return ImageLabels(labs, obj_table=obj_table)

    @staticmethod
    def from_file(f: Union[str, pathlib.Path, IO]) -> 'ImageLabels':
        if isinstance(f, str):
            f = pathlib.Path(f).open('r')
        elif isinstance(f, pathlib.Path):
            f = f.open('r')
        elif isinstance(f, io.IOBase):
            pass
        else:
            raise TypeError('f should be a path as a string or `pathlib.Path` or a file, not a {}'.format(type(f)))
        return ImageLabels.from_json(json.load(f))

    @classmethod
    def from_contours(cls, label_contours: Sequence[Sequence[np.ndarray]],
                      label_classes: Optional[Union[str, Sequence[str]]] = None,
                      sources: Optional[Union[str, Sequence[str]]] = None,
                      id_prefix: Optional[str] = None) -> 'ImageLabels':
        """
        Convert a list of contours to an `ImageLabels` instance.

        :param label_contours: list of list of contours. The outer list contains one item per label,
                with the inner lists containing the contours that should be grouped to form a label.
                Each contour is an `(N, (y, x))` numpy array.
                If an inner list contains one contour, a polygonal label is created to represent it.
                If it contains more than one, a polygonal label is created for each member contour
                and they are combined with a group.
        :param label_classes: [optional] a list of the same length as `list_of_contours` that provides
                the label class of each contour, or a string to assign the same class to every label
        :param sources: [optional] a list of the same length as `list_of_contours` that provides
                the source of each contour, or a string to assign the same source to every label
        :return: an `ImageLabels` instance containing the labels extracted from the contours
        """
        if id_prefix is None:
            id_prefix = str(uuid.uuid4())
        obj_table = ObjectTable(id_prefix=id_prefix)
        labels = []
        if isinstance(label_classes, str) or label_classes is None:
            label_classes = itertools.repeat(label_classes)
        if isinstance(sources, str) or sources is None:
            sources = itertools.repeat(sources)
        for contours_in_label, lcls, lsrc in zip(label_contours, label_classes, sources):
            regions = [contour[:, ::-1] for contour in contours_in_label]
            poly = PolygonLabel(regions, classification=lcls, source=lsrc)
            obj_table.register(poly)
            labels.append(poly)

        return cls(labels, obj_table=obj_table)


    @staticmethod
    def _get_label_meta(meta: Optional[Union[str, Mapping, Sequence]], label_i) -> Optional[Any]:
        """
        Get label metadata from a metadata mapping:
        - if `meta` is None, will return `None`
        - if `meta` is a str, will return `meta`
        - if `meta` is a dict, will return `meta.get(label_i)`
        - if `meta` is a list, will return `meta[label_i]`

        :param meta: metadata mapping
        :param label_i: label index
        :return: label metadata
        """
        if meta is None or isinstance(meta, str):
            return meta
        elif isinstance(meta, Mapping):
            return meta.get(label_i)
        elif isinstance(meta, Sequence):
            return meta[label_i]
        else:
            raise TypeError('should be None, str, dict or list, not a {}'.format(type(meta)))

    @classmethod
    def from_label_image(cls, labels: np.ndarray, label_classes: Optional[Union[str, Sequence[str]]] = None,
                         sources: Optional[Union[str, Sequence[str]]] = None,
                         return_label_indices: bool = False) -> Union['ImageLabels', Tuple['ImageLabels', List[int]]]:
        """
        Convert a integer label image to an `ImageLabels` instance.

        Converts label images to contours using Scikit-Image `find_contours`.

        :param labels: a `(h,w)` numpy array of dtype `int32` that gives an integer label for each
                pixel in the image. Label values start at 1; pixels with a value of 0 will not be
                included in the returned labels.
        :param label_classes: [optional] either:
                - a list that provides the class of each integer label (element 0 will be ignored)
                - a dict that maps label index to class
                - a string to assign the same class to every label
        :param sources: [optional] provides label sources; has the same format as `label_classes`
        :param return_label_indices: (default False) if True, return the index of the label used for each label
            in the returned `ImageLabels`
        :return: an `ImageLabels` instance containing the labels extracted from the label mask image, or
            a tuple of `(image_labels, label_indices)` where `image_labels` is an `ImageLabels` instance and
            `label_indices` gives the label index for each vectorized label in `image_labels`

        """
        if label_classes is not None and not isinstance(label_classes, (str, dict, list)):
            raise TypeError('label_classes should be None, a str, a dict or a list, not {}'.format(type(label_classes)))
        if sources is not None and not isinstance(sources, (str, dict, list)):
            raise TypeError('sources should be None, a str, a dict or a list, not {}'.format(type(sources)))
        contours = []
        lcls = []
        lsrc = []
        label_indices = []
        n_labels = labels.max()
        for i in range(1, n_labels+1):
            lmask = labels == i

            if lmask.sum() > 0:
                mask_positions = np.argwhere(lmask)
                (ystart, xstart), (ystop, xstop) = mask_positions.min(0), mask_positions.max(0) + 1

                if ystop >= ystart+1 and xstop >= xstart+1:
                    mask_trim = lmask[ystart:ystop, xstart:xstop]
                    mask_trim = np.pad(mask_trim, [(1,1), (1,1)], mode='constant').astype(np.float32)
                    cs = find_contours(mask_trim, 0.5)
                    regions = []
                    for contour in cs:
                        simp = _simplify_contour(contour + np.array((ystart, xstart)) - np.array([[1.0, 1.0]]))
                        if simp is not None:
                            regions.append(simp)
                    contours.append(regions)
                    lcls.append(cls._get_label_meta(label_classes, i))
                    lsrc.append(cls._get_label_meta(sources, i))
                    label_indices.append(i)

        img_labels = cls.from_contours(contours, lcls, lsrc)

        if return_label_indices:
            return img_labels, label_indices
        else:
            return img_labels

    @staticmethod
    def _contour_areas(contours: Sequence[np.ndarray]) -> np.ndarray:
        contour_areas = []
        for contour in contours:
            # Vectors from vertex 0 to all others
            u = contour[1:, :] - contour[0:1, :]
            contour_area = np.cross(u[:-1, :], u[1:, :]).sum() / 2
            contour_area = abs(float(contour_area))
            contour_areas.append(contour_area)
        return np.array(contour_areas)

    @classmethod
    def from_mask_images_cv(cls, masks: Sequence[np.ndarray],
                            label_classes: Optional[Union[str, Sequence[str]]] = None,
                            sources: Optional[Union[str, Sequence[str]]] = None, sort_decreasing_area: bool = True,
                            return_mask_indices: bool = False) -> \
            Union['ImageLabels', Tuple['ImageLabels', List[int]]]:
        """
        Convert labels represented as a sequence of mask images to an `ImageLabels` instance.
        Mask to contour conversion performed using OpenCV `findContours`, finding external contours only.

        Raises RuntimeError is OpenCV is not available.

        :param masks: a sequence of mask images - can be a generator or a 3D array. Each mask is a `(H,W)` array
            with non-zero values indicating pixels that are part of the label
        :param label_classes: [optional] either:
                - a list that provides the class of each mask
                - a dict that maps mask index to class
                - a string to assign the same class to every label
        :param sources: [optional] provides label sources; has the same format as `label_classes`
        :param sort_decreasing_area: (default True) if True, sort regions and labels in order of decreasing area
        :param return_mask_indices: (default False) if True, return the index of the mask used for each label
        :return: an `ImageLabels` instance, or
            a tuple of `(image_labels, mask_indices)` where `image_labels` is an `ImageLabels` instance and
            `mask_indices` gives the mask index for each vectorized label in `image_labels`
        """
        if cv2 is None:
            raise RuntimeError('OpenCV is not available!')
        if label_classes is not None and not isinstance(label_classes, (str, dict, list)):
            raise TypeError('label_classes should be None or a str, dict or list, not a {}'.format(type(label_classes)))
        if sources is not None and not isinstance(sources, (str, dict, list)):
            raise TypeError('sources should be None or a str, dict or list, not a {}'.format(type(sources)))

        mask_areas = []
        contours_classes_sources = []
        mask_indices = []
        n_masks = 0
        for mask_i, lab_msk in enumerate(masks):
            result = cv2.findContours((lab_msk != 0).astype(np.uint8), cv2.RETR_LIST,
                                      cv2.CHAIN_APPROX_TC89_L1)
            if len(result) == 3:
                _, region_contours, _ = result
            else:
                region_contours, _ = result
            
            region_contours = [contour[:, 0, ::-1] for contour in region_contours if len(contour) >= 3]

            if len(region_contours) > 0:
                # Compute area
                areas = cls._contour_areas(region_contours)
                mask_areas.append(float(areas.sum()))

                if sort_decreasing_area:
                    # Sort in order of decreasing area
                    order = np.argsort(areas)[::-1]
                    region_contours = [region_contours[i] for i in order]

                mask_indices.append(mask_i)

                contours_classes_sources.append((region_contours,
                                                 cls._get_label_meta(label_classes, mask_i),
                                                 cls._get_label_meta(sources, mask_i)))

            n_masks += 1
        mask_areas = np.array(mask_areas)

        if sort_decreasing_area and len(contours_classes_sources) > 0:
            order = np.argsort(mask_areas)[::-1]
            contours_classes_sources = [contours_classes_sources[i] for i in order]

        if len(contours_classes_sources) > 0:
            image_contours, lcls, lsrc = list(zip(*contours_classes_sources))
            img_labels = cls.from_contours(image_contours, lcls, lsrc)
        else:
            img_labels = cls.from_contours([])

        if return_mask_indices:
            return img_labels, mask_indices
        else:
            return img_labels


_INT_ID_PAT = re.compile('\d+')

def _generic_obj_id_update_helper(labels_json, id_prefix, id_remapping, idx_counter_in_list):
    modified = False
    if isinstance(labels_json, list):
        for x in labels_json:
            m = _generic_obj_id_update_helper(x, id_prefix, id_remapping, idx_counter_in_list)
            modified = modified or m
    elif isinstance(labels_json, dict):
        if 'label_type' in labels_json and 'label_class' in labels_json:
            # Its a label
            obj_id = labels_json.get('object_id')
            if obj_id is None:
                obj_id = '{}__{}'.format(id_prefix, idx_counter_in_list[0])
                idx_counter_in_list[0] += 1
                labels_json['object_id'] = obj_id
                modified = True
            if isinstance(obj_id, int):
                new_id = '{}__{}'.format(id_prefix, obj_id)
                labels_json['object_id'] = new_id
                id_remapping[obj_id] = new_id
                modified = True
            elif isinstance(obj_id, str):
                match = _INT_ID_PAT.match(obj_id)
                if match is not None and match.end(0) == len(obj_id):
                    new_id = '{}__{}'.format(id_prefix, obj_id)
                    labels_json['object_id'] = new_id
                    id_remapping[obj_id] = new_id
                    modified = True

        for x in labels_json.values():
            m = _generic_obj_id_update_helper(x, id_prefix, id_remapping, idx_counter_in_list)
            modified = modified or m

    return modified


def _composite_obj_id_update_helper(labels_json, id_remapping):
    modified = False
    if isinstance(labels_json, list):
        for x in labels_json:
            m = _composite_obj_id_update_helper(x, id_remapping)
            modified = modified or m
    elif isinstance(labels_json, dict):
        if 'label_type' in labels_json and 'label_class' in labels_json:
            # Its a label;
            if labels_json['label_type'] == 'composite':
                component_ids = labels_json['components']
                new_component_ids = [id_remapping.get(x, x) for x in component_ids]
                if new_component_ids != component_ids:
                    modified = True
                    labels_json['components'] = new_component_ids

        for x in labels_json.values():
            m = _composite_obj_id_update_helper(x, id_remapping)
            modified = modified or m

    return modified


def ensure_json_object_ids_have_prefix(labels_json, id_prefix, id_remapping=None, idx_counter_in_list=None):
    """
    Fix JSON label representation, by prefixing object IDs with `id_prefix`.

    Modifies JSON data in place.

    Exmaple:
    >>> import uuid
    >>> id_prefix = str(uuid.uuid4())
    >>> ensure_json_object_ids_have_prefix(labels_json, id_prefix)

    :param labels_json: labels in JSON form
    :param id_prefix: object ID prefix
    :return: True if JSON data was changed
    """
    if id_prefix.strip() == '':
        raise ValueError('id_prefix should not be empty or whitespace')
    idx_counter_in_list = [1]
    id_remapping = {}
    m1 = _generic_obj_id_update_helper(labels_json, id_prefix, id_remapping, idx_counter_in_list)
    m2 = _composite_obj_id_update_helper(labels_json, id_remapping)
    return m1 or m2


class WrappedImageLabels:
    def __init__(self, image_filename: Optional[str] = None, completed_tasks: Optional[Container[str]] = None,
                 metadata: Optional[Dict] = None, labels_json: Optional[Any] = None,
                 labels: Optional[ImageLabels] = None):
        """
        :param image_filename: the image filename as a string
        :param completed_tasks: a list of completed tasks
        :param metadata: metadata as a dictionary
        :param labels_json: labels in JSON form (labels parameter should be None if labels_json provided)
        :param labels: labels as an `ImageLabels` instance (labels_json parameter should be None if labels provided)
        """
        if labels_json is None and labels is None:
            raise ValueError('Either labels_json should be provided or image_labels should be provided')
        if labels_json is not None and labels is not None:
            raise ValueError('Either labels_json should be provided or image_labels should be provided, not both')
        if metadata is None:
            metadata = {}
        if completed_tasks is None:
            completed_tasks = []
        self.image_filename = image_filename
        self.completed_tasks = completed_tasks
        self.metadata = metadata
        self.__labels_json = labels_json
        self.__labels = labels

    @property
    def labels(self) -> ImageLabels:
        if self.__labels is None:
            self.__labels = ImageLabels.from_json(self.__labels_json)
            self.__labels_json = None
        return self.__labels

    @labels.setter
    def labels(self, lab: ImageLabels):
        self.__labels = lab
        self.__labels_json = None

    @property
    def labels_json(self) -> Any:
        if self.__labels_json is not None:
            return self.__labels_json
        elif self.__labels is not None:
            return self.__labels.to_json()
        else:
            raise RuntimeError

    @labels_json.setter
    def labels_json(self, js: Any):
        self.__labels_json = js
        self.__labels = None

    @property
    def is_blank(self) -> bool:
        """Determine if these wrapped labels are blank. If they are blank, then there is no data store, so
        e.g. a labels file can be deleted. A `WrappedLabelsInstance` is considered blank if no
        labels/annotations are defined and no tasks have been marked as completed.

        :return: a boolean indicating if this instance is blank
        """
        if self.__labels_json is not None and self.__labels_json != []:
            return False
        elif self.__labels is not None and len(self.__labels) == 0:
            return False

        # Labels are empty, but if a task is marked as complete this indicates
        # that the image contains nothing of interest
        if len(self.completed_tasks) > 0:
            return False

        return True

    def with_labels(self, labels: ImageLabels) -> 'WrappedImageLabels':
        """Create a new WrappedLabels instance, replacing the labels

        :param labels: replacement `ImageLabels` instance
        :return: new `WrappedLabels` instance
        """
        return WrappedImageLabels(image_filename=self.image_filename, completed_tasks=self.completed_tasks,
                                  metadata=self.metadata, labels=labels)

    def to_json(self) -> Any:
        js = self.metadata.copy()
        if self.image_filename is not None:
            js['image_filename'] = self.image_filename
        js.update({'completed_tasks': list(self.completed_tasks),
                   'labels': self.labels_json})
        return js

    def write_to_file(self, f: IO):
        """Write labels to file

        :param f: a file-like object, or a path as a `str` or `pathlib.File`
        """
        if isinstance(f, str):
            f = pathlib.Path(f).open('w')
        elif isinstance(f, pathlib.Path):
            f = f.open('w')
        elif isinstance(f, io.IOBase):
            pass
        else:
            raise TypeError('f should be a path as a string or `pathlib.Path` or a file, not a {}'.format(type(f)))
        json.dump(self.to_json(), f)

    @staticmethod
    def from_json(js: Any) -> 'WrappedImageLabels':
        if isinstance(js, dict):
            metadata = js.copy()
            completed_tasks = []
            image_filename = None
            if 'complete' in js:
                # Old-style complete flag; transform to 'finished' task
                del metadata['complete']
                if js['complete']:
                    completed_tasks = ['finished']
                else:
                    completed_tasks = []
            if 'completed_tasks' in js:
                del metadata['completed_tasks']
                completed_tasks = js['completed_tasks']
            if 'image_filename' in js:
                del metadata['image_filename']
                image_filename = js['image_filename']
            del metadata['labels']
            return WrappedImageLabels(image_filename=image_filename, completed_tasks=completed_tasks,
                                      metadata=metadata, labels_json=js['labels'])
        elif isinstance(js, list):
            # No metadata: use defaults
            return WrappedImageLabels(labels_json=js)
        else:
            raise TypeError('Labels loaded from file must either be a dict or a list, '
                            'not a {0}'.format(type(js)))

    @staticmethod
    def from_file(f: Union[str, pathlib.Path, IO]) -> 'WrappedImageLabels':
        if isinstance(f, str):
            f = pathlib.Path(f)
        if isinstance(f, pathlib.Path):
            file = f.open('r')
        elif isinstance(f, io.IOBase):
            file = f
        else:
            raise TypeError('f should be a path as a string or `pathlib.Path` or a file, not a {}'.format(type(f)))
        js = json.load(file)
        return WrappedImageLabels.from_json(js)


@deprecated(reason='Please use labelled_image.LabelledImage.in_memory()')
def InMemoryLabelledImage(pixels, labels=None, completed_tasks=None):
    from image_labelling_tool.labelled_image import LabelledImage
    return LabelledImage.in_memory(pixels, WrappedImageLabels(labels=labels, completed_tasks=completed_tasks))


@deprecated(reason='Please use labelled_image.LabelledImage.for_image_label_file_pair()')
def PersistentLabelledImage(image_path, labels_path, readonly=False):
    from image_labelling_tool.labelled_image import LabelledImage
    return LabelledImage.for_image_label_file_pair(image_path, labels_path, readonly=readonly)


@deprecated(reason='Please use labelled_image.LabelledImage.for_directory()')
def _PersistentLabelledImage__for_directory(dir_path, image_filename_patterns=None, with_labels_only=False,
                                            labels_dir=None, readonly=False):
    from image_labelling_tool.labelled_image import LabelledImage
    if image_filename_patterns is None:
        image_filename_patterns = ['*.png']
    return LabelledImage.for_directory(dir_path, image_filename_patterns, with_labels_only=with_labels_only,
                                       labels_dir=labels_dir, readonly=readonly)

PersistentLabelledImage.for_directory = _PersistentLabelledImage__for_directory


@deprecated(reason='Please use labelled_image.LabelledImage.for_image_files()')
def _PersistentLabelledImage__for_files(image_paths, with_labels_only=False, labels_dir=None, readonly=False):
    from image_labelling_tool.labelled_image import LabelledImage
    return LabelledImage.for_image_files(image_paths, with_labels_only=with_labels_only,
                                         labels_dir=labels_dir, readonly=readonly)

PersistentLabelledImage.for_files = _PersistentLabelledImage__for_files


@deprecated(reason='Please use the FileImageSource, InMemoryLabelsStore and LabelledImage classes in '
                   'the labelled_image module. Please see the source code of the `LabelledImageFile` '
                   'function in the labelling_tool module to see how.')
def LabelledImageFile(path, labels=None, tasks_complete=None, on_set_labels=None):
    from image_labelling_tool.labelled_image import LabelledImage, FileImageSource, InMemoryLabelsStore

    if on_set_labels is not None:
        # The old callback expects an ImageLabels instance, so wrap it
        def on_update(wrapped_labels):
            on_set_labels(wrapped_labels.labels)
    else:
        on_update = None

    image_source = FileImageSource(path)
    if labels is None:
        labels = ImageLabels([])
    labels_store = InMemoryLabelsStore(WrappedImageLabels(completed_tasks=tasks_complete, labels=labels),
                                       on_update=on_update)

    return LabelledImage(image_source, labels_store)



"""Labelled image interfaces for Flask and Qt-based applications.

The classes implemented here provide the interface through which the Flask and Qt based labelling tools
access and update images and their labels.

The most convenient way of providing images to a Flask or Qt based labeler is using a list of `LabelledImage`
instances. Each `LabelledImage` has an image source and a labels store that provide the image and store
the labels respectively.
"""
import pathlib
import io
from abc import abstractmethod
import mimetypes
import json
import numpy as np
from typing import Union, Tuple, List, Any, Optional, Sequence, Mapping, Container, Callable
from PIL import Image
from skimage.util import img_as_ubyte
from skimage import transform
from image_labelling_tool.labelling_tool import WrappedImageLabels, ImageLabels, LabelClass


PathType = Union[pathlib.Path, str]


class ImageSource:
    """Image source abstract base class.

    Provides methods for accessing:
    - a local path on disk, if available or None
    - binary data and mime type, where the image format is a format supported by a web browser
    - image as either a NumPy array or a Pillow image, whichever is more convenient
    - the image size as a `(height, width)` tuple

    These methods are used by the Flask and Qt based labelers to retrieve images.
    """
    @property
    @abstractmethod
    def local_path(self) -> Optional[pathlib.Path]:
        """Get the path of the image on disk, if available, otherwise return None
        :return: path as a string or `None`
        """
        pass

    @abstractmethod
    def image_binary_and_mime_type(self) -> Tuple[bytes, str]:
        """Get the image as a tuple of its binary data and mime type
        :return: tuple `(binary_buffer, mime_type)`
        """
        pass

    @abstractmethod
    def image_as_array_or_pil(self) -> Union[np.ndarray, Image.Image]:
        """Get the image as either a NumPy array or a `PIL.Image`
        :return: NumPy array or `PIL.Image`
        """
        pass

    @property
    @abstractmethod
    def image_size(self) -> Tuple[int, int]:
        """Get the image size as a `(height, width)` tuple
        :return: `(height, width)`
        """
        pass


class FileImageSource (ImageSource):
    class MRUImageCache:
        def __init__(self, size: int = 32):
            """Image cache retaining most recently used entries

            :param size: the maximum number of images to store
            """
            self._path_to_image = {}
            self._access_order = []
            self._size = size

        def __call__(self, path: pathlib.Path):
            path_str = str(path.absolute())
            if path_str in self._path_to_image:
                self._access_order.remove(path_str)
                self._access_order.append(path_str)
                return self._path_to_image[path_str]
            else:
                # Trim cache first
                trim = self._size - 1
                to_remove = self._access_order[:-trim]
                for r in to_remove:
                    del self._path_to_image[r]
                del self._access_order[:-trim]
                # Load image and cache
                img = Image.open(path_str)
                self._path_to_image[path_str] = img
                self._access_order.append(path_str)
                return img

    def __init__(self, image_path: PathType, image_loader: Any = None, store_locally: bool = False):
        """Local file image source.
        The image is stored as a file on disk.

        :param image_path: the path at which the image can be found as a `str` or `pathlib.Path`
        :param image_loader: [optional] a function that loads images to be returned by the `image_for_dextr`
            method. If you want to cache the images to avoid loading them each time, but limit the
            number of images kept in memory, use a `LocalFileImageSource.MRUImageCache` instance.
            Alternatively, the `store_locally` parameter caches it with the instance
        :param store_locally: [optionally] if True, when `image_for_dextr` is called the image will be stored
            in this instance. This could consume a lot of memory if many `LocalFileImageSource` instances
            have their `image_for_dextr` methods called. Note that if `image_loader` is provided and
            `store_locally` is True, then the image will be stored on this instance as well.
        """
        if isinstance(image_path, str):
            image_path = pathlib.Path(image_path)
        self.image_path = image_path
        self._image_loader = image_loader
        self._store_locally = store_locally
        self.__image = None
        self.__image_size = None

    @property
    def local_path(self) -> Optional[pathlib.Path]:
        return self.image_path

    def image_binary_and_mime_type(self) -> Tuple[bytes, str]:
        data = self.image_path.open('rb').read()
        mime_type = mimetypes.guess_type(self.image_path)[0]
        return data, mime_type

    def image_as_array_or_pil(self) -> Union[np.ndarray, Image.Image]:
        """Get the image as either a NumPy array or a `PIL.Image`
        :return: NumPy array or `PIL.Image`
        """
        if self._store_locally and self.__image is not None:
            # Stored locally from a previous call
            return self.__image

        # Get the image
        if self._image_loader is not None:
            img = self._image_loader(self.image_path)
        else:
            img = Image.open(str(self.image_path))

        # Store locally if needed
        if self._store_locally:
            self.__image = img

        # Store the image size
        self.__image_size = img.size[::-1]

        return img

    @property
    def image_size(self) -> Tuple[int, int]:
        """Get the image size as a `(height, width)` tuple
        :return: `(height, width)`
        """
        if self.__image_size is None:
            # Get the image size
            self.__image_size = Image.open(str(self.image_path)).size[::-1]
        return self.__image_size


class InMemoryImageSource (ImageSource):
    def __init__(self, image: Union[np.ndarray, Image.Image]):
        """In memory image source, where the image is represented as either a NumPy array or a PIL Image.

        :param image: image as a NumPy array or PIL Image
        """
        self.image = image

    @property
    def local_path(self) -> Optional[pathlib.Path]:
        return None

    def image_binary_and_mime_type(self) -> Tuple[bytes, str]:
        buf = io.BytesIO()
        if isinstance(self.image, np.ndarray):
            # Convert NumPy array to PIL Image
            pix_u8 = img_as_ubyte(self.image)
            img = Image.fromarray(pix_u8)
        elif isinstance(self.image, Image.Image):
            img = self.image
        else:
            raise TypeError('image is neither a np.ndarray or a PIL Image')
        img.save(buf, format='bmp')
        return buf.getvalue(), 'image/bmp'

    def image_as_array_or_pil(self) -> Union[np.ndarray, Image.Image]:
        """Get the image as either a NumPy array or a `PIL.Image`
        :return: NumPy array or `PIL.Image`
        """
        return self.image

    @property
    def image_size(self) -> Tuple[int, int]:
        """Get the image size as a `(height, width)` tuple
        :return: `(height, width)`
        """
        if isinstance(self.image, np.ndarray):
            return self.image.shape[:2]
        elif isinstance(self.image, Image.Image):
            return self.image.size[::-1]
        else:
            raise TypeError('image is neither a np.ndarray or a PIL Image')


class LabelsStore:
    """Labels store abstract base class.

    Provides the interface by which the labelling tool accesses and updates labels.

    Labels take the form of `WrappedImageLabels` instances that also keep metadata such as the list
    of completed tasks, etc.
    """
    @abstractmethod
    def get_wrapped_labels(self) -> WrappedImageLabels:
        """Get the labels as a `WrappedImageLabels` instance.
        :return: labels as a `WrappedImageLabels` instance.
        """
        pass

    @abstractmethod
    def update_wrapped_labels(self, wrapped_labels: WrappedImageLabels):
        """Update the labels

        :param wrapped_labels: updated wrapped labels
        """
        pass

    @property
    def has_labels(self) -> bool:
        """Return True if labels exist

        :return: boolean
        """
        return False


class FileLabelsStore (LabelsStore):
    def __init__(self, labels_path: PathType, image_filename: Optional[str] = None,
                 readonly: bool = False, delete_if_blank: bool = True):
        """Labels stored in a file on disk.

        Getting the labels will load them from the file, while updating them (e.g. in response
        to user actions within the labelling tool) will write the labels to the file.

        :param labels_path: labels file path as a `str` or `pathlib.Path`
        :param image_filename: [optional] the image filename to be stored in the label metadata,
            only used in cases where the file at `labels_path` does not exist and a blank
            `WrappedImageLabels` instance needs to be created
        :param readonly: if True, updating the labels via the `update_wrapped_labels` method will
            not write the labels to the file at `labels_path`
        :param delete_if_blank: if True, the labels file at `labels_path` will be deleted if
            the `update_wrapped_labels` method is called with blank labels (no labels defined,
            no completed tasks). `update_wrapped_labels` is normally invoked due to user actions
            from within the tool.
        """
        if isinstance(labels_path, str):
            labels_path = pathlib.Path(labels_path)

        self.labels_path = labels_path
        self.image_filename = image_filename
        self.readonly = readonly
        self.delete_if_blank = delete_if_blank
        self._labels = None

    def get_wrapped_labels(self) -> WrappedImageLabels:
        if self._labels is None:
            if self.labels_path.exists():
                self._labels = WrappedImageLabels.from_file(self.labels_path)
            else:
                self._labels = WrappedImageLabels(
                    image_filename=self.image_filename, labels=ImageLabels([]))
        return self._labels

    def update_wrapped_labels(self, wrapped_labels: WrappedImageLabels):
        """Update the labels

        :param wrapped_labels: updated wrapped labels
        """
        self._labels = wrapped_labels
        if not self.readonly:
            if wrapped_labels.is_blank and self.delete_if_blank:
                # Blank and deletion is requested
                if self.labels_path.exists():
                    self.labels_path.unlink()
            else:
                with self.labels_path.open('w') as f_out:
                    json.dump(wrapped_labels.to_json(), f_out, indent=2)

    @property
    def has_labels(self) -> bool:
        """Return True if labels exist

        :return: boolean
        """
        return self.labels_path.exists()


class InMemoryLabelsStore (LabelsStore):
    def __init__(self, wrapped_labels: WrappedImageLabels = None, image_filename: str = None,
                 on_update: Optional[Callable[[WrappedImageLabels], None]] = None):
        """Labels stored in memory.

        The labels are kept in memory as a `WrappedImageLabels` instance.

        :param wrapped_labels: [optional] the initial labels at the start
        :param image_filename: [optional] the image filename to be stored in the label metadata,
            only used in cases where the file at `labels_path` does not exist and a blank
            `WrappedImageLabels` instance needs to be created
        :param on_update: [optional] a callback function of the form `fn(wrapped_labels)` that will be
            invoked when the labels are updated, normally due to user actions
        """
        if wrapped_labels is None:
            wrapped_labels = WrappedImageLabels(
                    image_filename=image_filename, labels=ImageLabels([]))
        self.wrapped_labels = wrapped_labels
        self._on_update = on_update

    def get_wrapped_labels(self) -> WrappedImageLabels:
        return self.wrapped_labels

    def update_wrapped_labels(self, wrapped_labels):
        """Update the labels

        :param wrapped_labels: updated wrapped labels
        """
        self.wrapped_labels = wrapped_labels
        if self._on_update is not None:
            self._on_update(self.wrapped_labels)

    @property
    def has_labels(self) -> bool:
        """Return True if labels exist

        :return: boolean
        """
        return not self.wrapped_labels.is_blank

    @staticmethod
    def from_json(labels_js: Any) -> 'InMemoryLabelsStore':
        """Construct an in-memory labels store from JSON representation

        :param labels_js: labels in JSON form
        :return: an `InMemoryLabelsStore`
        """
        wrapped_labels = WrappedImageLabels.from_json(labels_js)
        return InMemoryLabelsStore(wrapped_labels)


_LabelClassMappingType = Union[Mapping[str, int],
                               Callable[[str], int],
                               Sequence[Union[str, LabelClass, Sequence[Union[str, LabelClass]]]]]


class LabelledImage:
    """A labelled image; pairs an image source (file/memory/etc.) with a labels store (file/memory/etc.)

    Provides convenience methods for warping the images and labels and rendering the labels.
    """
    def __init__(self, image_source: ImageSource, labels_store: LabelsStore):
        """A labelled image.

        The interface through which labelling tools can conveniently access images to label and
        access and update their corresponding labels.

        An image source that provides the image is paired with a labels store that stores the labels.

        You may combine on-disk or in-memory images and labels using the classes defined in this module.
        Alternatively you can derive from the abstract base classes in order to provide your own.

        :param image_source: an `ImageSource` instance
        :param labels_store: a `LabelsStore` instance
        """
        self.image_source = image_source
        self.labels_store = labels_store

    @property
    def labels(self) -> ImageLabels:
        return self.labels_store.get_wrapped_labels().labels

    @property
    def wrapped_labels(self) -> WrappedImageLabels:
        return self.labels_store.get_wrapped_labels()

    @property
    def has_labels(self) -> bool:
        return self.labels_store.has_labels

    @property
    def wrapped_labels(self) -> WrappedImageLabels:
        return self.labels_store.get_wrapped_labels()

    def warped(self, transformation: Any, sz_px: Tuple[int, int]) -> 'LabelledImage':
        """Warp the image and associated labels

        :param transformation: transformation to apply as a `skimage.transform._geometric.GeometricTransform`
        :param sz_px: the size of the transformed image in pixels as a `(height, width)` tuple
        :return:
        """
        # Get the image as a NumPy array
        image_pixels = np.array(self.image_source.image_as_array_or_pil())
        # Warp the image
        warped_pixels = transform.warp(image_pixels, transformation.inverse)[:int(sz_px[0]), :int(sz_px[1]), ...]
        # Get the wrapped labels
        wlabels = self.labels_store.get_wrapped_labels()
        # Get the ImageLabels instance and warp it
        warped_image_labels = wlabels.labels.warped(transformation)
        # Wrapped labels instance with labels replaced
        warped_labels = wlabels.with_labels(warped_image_labels)
        return LabelledImage(InMemoryImageSource(warped_pixels), InMemoryLabelsStore(warped_labels))

    def render_label_classes(self, label_classes: _LabelClassMappingType,
                             multichannel_mask: bool = False, fill: bool = True) -> np.ndarray:
        """
        Render label classes to a create a label class image suitable for use as a
        semantic segmentation ground truth image.

        See `ImageLabels.render_label_classes` for full description and parameter documentation.

        :param label_classes: label class mapping
        :param multichannel_mask: if False return label class image, if True return multi-channel mask
        :param fill: if True, labels will be filled, otherwise they will be outlined
        :return: (H,W) array with dtype=int if multichannel is False, otherwise (H,W,n_classes) with dtype=bool
        """
        return self.labels_store.get_wrapped_labels().labels.render_label_classes(
            label_classes, self.image_source.image_size, multichannel_mask=multichannel_mask, fill=fill)

    def render_label_instances(self, label_classes: Optional[_LabelClassMappingType],
                               multichannel_mask: bool = False,
                               fill: bool = True) -> Union[Tuple[np.ndarray, np.ndarray],
                                                           Tuple[np.ndarray, np.ndarray, np.ndarray]]:
        """
        Render a label instance image suitable for use as an instance segmentation ground truth image.

        To get a stack of masks with one mask per object/label, give `multichannel_mask` a value of True

        :param label_classes: [optional] label class mapping, if None render all labels
        :param multichannel_mask: if False return label class image, if True return multi-channel mask
        :param fill: if True, labels will be filled, otherwise they will be outlined
        :return: tuple of (label_image, label_index_to_cls) where:
            label_image is a (H,W) array with dtype=int
            label_index_to_cls is a 1D array that gives the class index of each labels. The first entry
                at index 0 will have a value of 0 as it is the background label. The class indices are the
                index of the class in `label_class` + 1.
        """
        return self.labels_store.get_wrapped_labels().labels.render_label_instances(
            label_classes, self.image_source.image_size, multichannel_mask=multichannel_mask, fill=fill)

    def extract_label_images(self, label_class_set: Optional[Container[str]] = None) -> List[np.ndarray]:
        """
        Extract an image of each labelled entity.
        The resulting image is the original image masked with an alpha channel that results from rendering the label

        :param label_class_set: a set/container of classes whose labels should be rendered, or None for all labels
        :return: a list of (H,W,C) image arrays
        """
        # Get the image as a NumPy array
        image_pixels = np.array(self.image_source.image_as_array_or_pil())
        return self.labels_store.get_wrapped_labels().labels.extract_label_images(
            image_pixels, label_class_set=label_class_set)

    @staticmethod
    def _compute_labels_path(image_path: pathlib.Path, labels_dir: pathlib.Path,
                             label_suffix: str) -> pathlib.Path:
        filename = image_path.stem + label_suffix
        if labels_dir is None:
            labels_dir = image_path.parent
        return labels_dir / filename

    @classmethod
    def for_directory(cls, images_dir: PathType, image_filename_patterns: Sequence[str] = ('*.png', '*.jpg'),
                      with_labels_only: bool = False, labels_dir: Optional[PathType] = None,
                      readonly: bool = False, delete_if_blank: bool = True,
                      label_suffix: str = '__labels.json', image_loader: Any = None,
                      store_locally: bool = False) -> List['LabelledImage']:
        """Search a directory for image files and construct `LabelledImage` instances for them.
        If `with_labels_only` is True, image files without corresponding label files will not be included.
        If `labels_dir` is provided, the label files will be stored in that directory, otherwise by default
        a label file will be stored in the same directory as its' corresponding image.
        Filenames for label files are generated by stripping the extension from the image filename and
        appending `label_suffix`; providing a value for this parameter allows you customize the suffix
        and extension.

        :param images_dir: the path of the directory to search as a `str` or `pathlib.Path` instance
        :param image_filename_patterns: sequence of image filename patterns, e.g. '*.png'. Default
            is `['*.png', '*.jpg']`, so all PNG and JPEG files will be used.
        :param with_labels_only: if True, skip images for which no corresponding label file can be found
        :param labels_dir: [optional] a directory as a `str` or `pathlib.Path` in which the label files are to
            be stored
        :param readonly: if True, labels are read-only and changes made by the user will *not* be saved to the
            label files
        :param delete_if_blank: if True, the labels files will be deleted if user actions in the tool
            result in blank labels (no labels defined and no completed tasks)
        :param label_suffix: the suffix for label files
        :param image_loader: image loader used for caching etc. (see `FileImageSource` constructor)
        :param store_locally: if True, the `FileImageSource` created for each image will cache
            the image in memory, if loaded for DEXTR, etc. (note that this could consume a lot of memory).
        :return: a list of `LabelledImage` instances
        """
        if isinstance(images_dir, str):
            images_dir = pathlib.Path(images_dir)
        images_dir = images_dir.absolute()

        image_paths = []
        for pat in image_filename_patterns:
            image_paths.extend(list(images_dir.glob(pat)))
        return cls.for_image_files(image_paths, with_labels_only=with_labels_only, labels_dir=labels_dir,
                                   readonly=readonly, delete_if_blank=delete_if_blank, label_suffix=label_suffix,
                                   image_loader=image_loader, store_locally=store_locally)

    @classmethod
    def for_image_files(cls, image_paths: Sequence[PathType], with_labels_only: bool = False,
                        labels_dir: Optional[PathType] = None, readonly: bool = False, delete_if_blank: bool = True,
                        label_suffix: str = '__labels.json', image_loader: Any = None,
                        store_locally: bool = False) -> List['LabelledImage']:
        """Construct `LabelledImage` instances for the on-disk image files in `image_paths`.
        If `with_labels_only` is True, image files without corresponding label files will not be included.
        If `labels_dir` is provided, the label files will be stored in that directory, otherwise by default
        a label file will be stored in the same directory as its' corresponding image.
        Filenames for label files are generated by stripping the extension from the image filename and
        appending `label_suffix`; providing a value for this parameter allows you customize the suffix
        and extension.

        Note that in the case where the label file path generation process would result in multiple images
        having the same labels path, a `ValueError` is raised.

        :param image_paths: the paths of the images as a sequence of `str` or `pathlib.Path` instances
        :param with_labels_only: if True, skip images for which no corresponding label file can be found
        :param labels_dir: [optional] a directory as a `str` or `pathlib.Path` in which the label files are to
            be stored
        :param readonly: if True, labels are read-only and changes made by the user will *not* be saved to the
            label files
        :param delete_if_blank: if True, the labels files will be deleted if user actions in the tool
            result in blank labels (no labels defined and no completed tasks)
        :param label_suffix: the suffix for label files
        :param image_loader: image loader used for caching etc. (see `FileImageSource` constructor)
        :param store_locally: if True, the `FileImageSource` created for each image will cache
            the image in memory, if loaded for DEXTR, etc. (note that this could consume a lot of memory).
        :return: a list of `LabelledImage` instances
        """
        if isinstance(labels_dir, str):
            labels_dir = pathlib.Path(labels_dir)

        path_pairs = []
        label_path_to_image_path = {}
        for img_path in image_paths:
            if isinstance(img_path, str):
                img_path = pathlib.Path(img_path)

            labels_path = cls._compute_labels_path(img_path, labels_dir=labels_dir, label_suffix=label_suffix)
            if labels_path in label_path_to_image_path:
                raise ValueError('Duplicate label paths: the images at {} and {} both have the '
                                 'same label path {}'.format(label_path_to_image_path[labels_path], img_path,
                                                             labels_path))

            path_pairs.append((img_path, labels_path))
            label_path_to_image_path[labels_path] = img_path

        return LabelledImage.for_image_label_file_pairs(path_pairs, with_labels_only=with_labels_only,
                                                        readonly=readonly, delete_if_blank=delete_if_blank,
                                                        image_loader=image_loader, store_locally=store_locally)

    @classmethod
    def for_image_label_file_pairs(cls, path_pairs: Sequence[Tuple[PathType, PathType]],
                                   with_labels_only: bool = False, readonly: bool = False,
                                   delete_if_blank: bool = True, image_loader: Any = None,
                                   store_locally: bool = False) -> List['LabelledImage']:
        """Construct `LabelledImage` instances for the on-disk image files in `image_paths`.
        If `with_labels_only` is True, image files without corresponding label files will not be included.

        :param path_pairs: a sequence of pairs (tuples) of `(image_path, labels_path)` where `images_path`
            and `labels_path` are `str` or `pathlib.Path` instances
        :param with_labels_only: if True, skip images for which no corresponding label file can be found
        :param readonly: if True, labels are read-only and changes made by the user will *not* be saved to the
            label files
        :param delete_if_blank: if True, the labels files will be deleted if user actions in the tool
            result in blank labels (no labels defined and no completed tasks)
        :param image_loader: image loader used for caching etc. (see `FileImageSource` constructor)
        :param store_locally: if True, the `FileImageSource` created for each image will cache
            the image in memory, if loaded for DEXTR, etc. (note that this could consume a lot of memory).
        :return: a list of `LabelledImage` instances
        """
        limgs = []
        for img_path, labels_path in path_pairs:
            if isinstance(img_path, str):
                img_path = pathlib.Path(img_path)
            if isinstance(labels_path, str):
                labels_path = pathlib.Path(labels_path)

            if not with_labels_only or labels_path.exists():
                limgs.append(cls.for_image_label_file_pair(img_path, labels_path, readonly=readonly,
                                                           delete_if_blank=delete_if_blank, image_loader=image_loader,
                                                           store_locally=store_locally))
        return limgs

    @classmethod
    def for_image_label_file_pair(cls, image_path: PathType, labels_path: PathType, readonly: bool = False,
                                  delete_if_blank: bool = True, image_loader: Any = None,
                                  store_locally: bool = False) -> 'LabelledImage':
        """Construct a `LabelledImage` instance for an on-disk image file paired with a label file.

        :param image_path: image path as a `str` or `pathlib.Path` instance
        :param labels_path: labels file path as a `str` or `pathlib.Path` instance
        :param readonly: if True, labels are read-only and changes made by the user will *not* be saved to the
            label files
        :param delete_if_blank: if True, the labels files will be deleted if user actions in the tool
            result in blank labels (no labels defined and no completed tasks)
        :param image_loader: image loader used for caching etc. (see `FileImageSource` constructor)
        :param store_locally: if True, the `FileImageSource` created for each image will cache
            the image in memory, if loaded for DEXTR, etc. (note that this could consume a lot of memory).
        :return: a `LabelledImage` instance
        """
        if isinstance(image_path, str):
            image_path = pathlib.Path(image_path)
        if isinstance(labels_path, str):
            labels_path = pathlib.Path(labels_path)

        image_source = FileImageSource(image_path, image_loader=image_loader, store_locally=store_locally)
        labels_store = FileLabelsStore(labels_path, image_filename=image_path.name,
                                       readonly=readonly, delete_if_blank=delete_if_blank)
        return LabelledImage(image_source, labels_store)

    @classmethod
    def in_memory(cls, image, wrapped_labels):
        image_source = InMemoryImageSource(image)
        labels_store = InMemoryLabelsStore(wrapped_labels)
        return LabelledImage(image_source, labels_store)


def shuffle_images_without_labels(labelled_images: Sequence[LabelledImage],
                                  random_state: Union[np.random.RandomState, int] = 12345) -> List[LabelledImage]:
    """Re-orderes a sequence if labelled images, such that those with labels come before those without,
    and the images without labels are shuffled.

    :param labelled_images: images to re-order
    :param random_state: [optional] random state for shuffling
    :return: re-ordered list of labelled images
    """
    if isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)
    with_labels = [img   for img in labelled_images   if img.has_labels]
    without_labels = [img   for img in labelled_images   if not img.has_labels]
    random_state.shuffle(without_labels)
    return with_labels + without_labels
